# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.piecewise.transform.piecewise_linear_transformation_base import (
    PiecewiseLinearTransformationBase,
)
from pyomo.core import Constraint, Binary, Var, RangeSet, Set
from pyomo.core.base import TransformationFactory
from pyomo.common.errors import DeveloperError
from math import ceil, log2


@TransformationFactory.register(
    "contrib.piecewise.disaggregated_logarithmic",
    doc="""
    Represent a piecewise linear function "logarithmically" by using a MIP with
    log_2(|P|) binary decision variables. This is a direct-to-MIP transformation;
    GDP is not used.
    """,
)
class DisaggregatedLogarithmicMIPTransformation(PiecewiseLinearTransformationBase):
    """Represent a piecewise linear function "logarithmically" as a MIP.

    This transformation represents a piecewise linear function
    "logarithmically" by using a MIP with :math:`log_2(|P|)` binary
    decision variables, following the "disaggregated logarithmic" method
    from [VAN10]_.

    This is a direct-to-MIP transformation; GDP is not used.  This
    method of logarithmically formulating the piecewise linear function
    imposes no restrictions on the family of polytopes, but we assume we
    have simplices in this code.

    """

    CONFIG = PiecewiseLinearTransformationBase.CONFIG()
    _transformation_name = "pw_linear_disaggregated_log"

    # Implement to use PiecewiseLinearTransformationBase. This function returns the Var
    # that replaces the transformed piecewise linear expr
    def _transform_pw_linear_expr(self, pw_expr, pw_linear_func, transformation_block):
        # Get a new Block for our transformation in transformation_block.transformed_functions,
        # which is a Block(Any). This is where we will put our new components.
        transBlock = transformation_block.transformed_functions[
            len(transformation_block.transformed_functions)
        ]

        # Dimensionality of the PWLF
        dimension = pw_expr.nargs()
        transBlock.dimension_indices = RangeSet(0, dimension - 1)

        # Substitute Var that will hold the value of the PWLE
        substitute_var = transBlock.substitute_var = Var()
        pw_linear_func.map_transformation_var(pw_expr, substitute_var)

        # Bounds for the substitute_var that we will widen
        substitute_var_lb = float("inf")
        substitute_var_ub = -float("inf")

        # Simplices are tuples of indices of points. Give them their own indices, too
        simplices = pw_linear_func._simplices
        num_simplices = len(simplices)
        transBlock.simplex_indices = RangeSet(0, num_simplices - 1)
        # Assumption: the simplices are really full-dimensional simplices and all have the
        # same number of points, which is dimension + 1
        transBlock.simplex_point_indices = RangeSet(0, dimension)

        # Enumeration of simplices: map from simplex number to simplex object
        idx_to_simplex = {k: v for k, v in zip(transBlock.simplex_indices, simplices)}

        # List of tuples of simplex indices with their linear function
        simplex_indices_and_lin_funcs = list(
            zip(transBlock.simplex_indices, pw_linear_func._linear_functions)
        )

        # We don't seem to get a convenient opportunity later, so let's just widen
        # the bounds here. All we need to do is go through the corners of each simplex.
        for P, linear_func in simplex_indices_and_lin_funcs:
            for v in transBlock.simplex_point_indices:
                val = linear_func(*pw_linear_func._points[idx_to_simplex[P][v]])
                if val < substitute_var_lb:
                    substitute_var_lb = val
                if val > substitute_var_ub:
                    substitute_var_ub = val
        transBlock.substitute_var.setlb(substitute_var_lb)
        transBlock.substitute_var.setub(substitute_var_ub)

        log_dimension = ceil(log2(num_simplices))
        transBlock.log_simplex_indices = RangeSet(0, log_dimension - 1)
        transBlock.binaries = Var(transBlock.log_simplex_indices, domain=Binary)

        # Injective function B: \mathcal{P} -> {0,1}^ceil(log_2(|P|)) used to identify simplices
        # (really just polytopes are required) with binary vectors. Any injective function
        # is enough here.
        B = {}
        for i in transBlock.simplex_indices:
            # map index(P) -> corresponding vector in {0, 1}^n
            B[i] = self._get_binary_vector(i, log_dimension)

        # Build up P_0 and P_plus ahead of time.

        # {P \in \mathcal{P} | B(P)_l = 0}
        def P_0_init(m, l):
            return [p for p in transBlock.simplex_indices if B[p][l] == 0]

        transBlock.P_0 = Set(transBlock.log_simplex_indices, initialize=P_0_init)

        # {P \in \mathcal{P} | B(P)_l = 1}
        def P_plus_init(m, l):
            return [p for p in transBlock.simplex_indices if B[p][l] == 1]

        transBlock.P_plus = Set(transBlock.log_simplex_indices, initialize=P_plus_init)

        # The lambda variables \lambda_{P,v} are indexed by the simplex and the point in it
        transBlock.lambdas = Var(
            transBlock.simplex_indices, transBlock.simplex_point_indices, bounds=(0, 1)
        )

        # Numbered citations are from Vielma et al 2010, Mixed-Integer Models
        # for Nonseparable Piecewise-Linear Optimization

        # Sum of all lambdas is one (6b)
        transBlock.convex_combo = Constraint(
            expr=sum(
                transBlock.lambdas[P, v]
                for P in transBlock.simplex_indices
                for v in transBlock.simplex_point_indices
            )
            == 1
        )

        # The branching rules, establishing using the binaries that only one simplex's lambda
        # coefficients may be nonzero
        # Enabling lambdas when binaries are on
        @transBlock.Constraint(transBlock.log_simplex_indices)  # (6c.1)
        def simplex_choice_1(b, l):
            return (
                sum(
                    transBlock.lambdas[P, v]
                    for P in transBlock.P_plus[l]
                    for v in transBlock.simplex_point_indices
                )
                <= transBlock.binaries[l]
            )

        # Disabling lambdas when binaries are on
        @transBlock.Constraint(transBlock.log_simplex_indices)  # (6c.2)
        def simplex_choice_2(b, l):
            return (
                sum(
                    transBlock.lambdas[P, v]
                    for P in transBlock.P_0[l]
                    for v in transBlock.simplex_point_indices
                )
                <= 1 - transBlock.binaries[l]
            )

        # for i, (simplex, pwlf) in enumerate(choices):
        # x_i = sum(lambda_P,v v_i, P in polytopes, v in V(P))
        @transBlock.Constraint(transBlock.dimension_indices)  # (6a.1)
        def x_constraint(b, i):
            return pw_expr.args[i] == sum(
                transBlock.lambdas[P, v]
                * pw_linear_func._points[idx_to_simplex[P][v]][i]
                for P in transBlock.simplex_indices
                for v in transBlock.simplex_point_indices
            )

        # Make the substitute Var equal the PWLE (6a.2)
        transBlock.set_substitute = Constraint(
            expr=substitute_var
            == sum(
                transBlock.lambdas[P, v]
                * linear_func(*pw_linear_func._points[idx_to_simplex[P][v]])
                for v in transBlock.simplex_point_indices
                for (P, linear_func) in simplex_indices_and_lin_funcs
            )
        )

        return substitute_var

    # Not a Gray code, just a regular binary representation
    # TODO test the Gray codes too
    # note: Must have num != 0 and ceil(log2(num)) > length to be valid
    def _get_binary_vector(self, num, length):
        ans = []
        for i in range(length):
            ans.append(num & 1)
            num >>= 1
        assert not num
        ans.reverse()
        return tuple(ans)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.version as pyomo_ver


class Tests(unittest.TestCase):
    def test_releaselevel(self):
        _relLevel = pyomo_ver.version_info[3].split('{')[0].strip()
        self.assertIn(_relLevel, ('devel', 'VOTD', 'final'))

    def test_version(self):
        try:
            from importlib.metadata import version

            pyomo_version = version('pyomo')
        except:
            self.skipTest('importlib.metadata is not available')

        if pyomo_ver.version_info[3] == 'final':
            self.assertEqual(pyomo_ver.version, pyomo_version)

        else:
            tmp_ = pyomo_version.split('.')
            self.assertEqual(str(tmp_[0]), str(pyomo_ver.version_info[0]))
            self.assertEqual(str(tmp_[1]), str(pyomo_ver.version_info[1]))
            if tmp_[-1].startswith('dev'):
                import pyomo.version.info as info

                self.assertEqual(int(tmp_[-1][3:]), info.serial)
                tmp_.pop()
            if len(tmp_) > 2:
                self.assertEqual(str(tmp_[2]), str(pyomo_ver.version_info[2]))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
import pyomo.contrib.mpc as mpc
from pyomo.contrib.mpc.examples.cstr.run_openloop import run_cstr_openloop

ipopt_available = pyo.SolverFactory("ipopt").available()


@unittest.skipIf(not ipopt_available, "ipopt is not available")
class TestCSTROpenLoop(unittest.TestCase):
    # This data was obtained from a run of this code. The test is
    # intended to make sure that values do not change, not that
    # they are correct in some absolute sense.
    _pred_A_data = [
        1.00000,
        1.50000,
        1.83333,
        2.05556,
        2.20370,
        2.30247,
        2.36831,
        2.41221,
        2.44147,
        2.46098,
        2.47399,
        2.48266,
        2.48844,
        2.36031,
        2.27039,
        2.20729,
        2.16301,
        2.13194,
        2.11013,
        2.09483,
        2.08409,
        2.07656,
        2.07127,
        2.06756,
        2.06495,
        2.16268,
        2.22893,
        2.27385,
        2.30431,
        2.32495,
        2.33895,
        2.34844,
        2.35488,
        2.35924,
        2.36220,
        2.36420,
        2.36556,
        2.36648,
        2.36711,
        2.36753,
        2.36782,
    ]
    _pred_B_data = [
        0.00000,
        0.30200,
        0.61027,
        0.90132,
        1.16380,
        1.39353,
        1.59049,
        1.75683,
        1.89576,
        2.01081,
        2.10544,
        2.18289,
        2.24600,
        2.41517,
        2.54001,
        2.63284,
        2.70242,
        2.75502,
        2.79516,
        2.82605,
        2.85007,
        2.86890,
        2.88380,
        2.89569,
        2.90526,
        2.81484,
        2.75455,
        2.71450,
        2.68802,
        2.67062,
        2.65927,
        2.65195,
        2.64728,
        2.64436,
        2.64258,
        2.64153,
        2.64096,
        2.64067,
        2.64057,
        2.64058,
        2.64064,
    ]

    def _get_input_sequence(self):
        input_sequence = mpc.TimeSeriesData(
            {"flow_in[*]": [0.1, 1.0, 0.7, 0.9]}, [0.0, 3.0, 6.0, 10.0]
        )
        return mpc.data.convert.series_to_interval(input_sequence)

    def test_openloop_simulation(self):
        input_sequence = self._get_input_sequence()
        ntfe = 4
        model_horizon = 1.0
        simulation_steps = 10
        m, sim_data = run_cstr_openloop(
            input_sequence, model_horizon=1.0, ntfe=4, simulation_steps=10
        )
        sim_time_points = [
            model_horizon / ntfe * i for i in range(simulation_steps * ntfe + 1)
        ]

        AB_data = sim_data.extract_variables([m.conc[:, "A"], m.conc[:, "B"]])

        A_cuid = sim_data.get_cuid(m.conc[:, "A"])
        B_cuid = sim_data.get_cuid(m.conc[:, "B"])
        pred_data = {A_cuid: self._pred_A_data, B_cuid: self._pred_B_data}

        self.assertStructuredAlmostEqual(pred_data, AB_data.get_data(), delta=1e-3)
        self.assertStructuredAlmostEqual(
            sim_time_points, AB_data.get_time_points(), delta=1e-7
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    log,
    log10,
    sin,
    cos,
    tan,
    sinh,
    cosh,
    tanh,
    asin,
    acos,
    atan,
    asinh,
    acosh,
    atanh,
    exp,
    sqrt,
    ceil,
    floor,
)
from math import e, pi

model = ConcreteModel()

# pick a value in the domain of all of these functions
model.ONE = Var(initialize=1)
model.ZERO = Var(initialize=0)


model.obj = Objective(expr=model.ONE + model.ZERO)

model.c_log = Constraint(expr=log(model.ONE) == 0)
model.c_log10 = Constraint(expr=log10(model.ONE) == 0)

model.c_sin = Constraint(expr=sin(model.ZERO) == 0)
model.c_cos = Constraint(expr=cos(model.ZERO) == 1)
model.c_tan = Constraint(expr=tan(model.ZERO) == 0)

model.c_sinh = Constraint(expr=sinh(model.ZERO) == 0)
model.c_cosh = Constraint(expr=cosh(model.ZERO) == 1)
model.c_tanh = Constraint(expr=tanh(model.ZERO) == 0)

model.c_asin = Constraint(expr=asin(model.ZERO) == 0)
model.c_acos = Constraint(expr=acos(model.ZERO) == pi / 2)
model.c_atan = Constraint(expr=atan(model.ZERO) == 0)

model.c_asinh = Constraint(expr=asinh(model.ZERO) == 0)
model.c_acosh = Constraint(expr=acosh((e**2 + model.ONE) / (2 * e)) == 0)
model.c_atanh = Constraint(expr=atanh(model.ZERO) == 0)

model.c_exp = Constraint(expr=exp(model.ZERO) == 1)
model.c_sqrt = Constraint(expr=sqrt(model.ONE) == 1)
model.c_ceil = Constraint(expr=ceil(model.ONE) == 1)
model.c_floor = Constraint(expr=floor(model.ONE) == 1)
model.c_abs = Constraint(expr=abs(model.ONE) == 1)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import (
    ConcreteModel,
    Param,
    Var,
    Objective,
    Constraint,
    RangeSet,
    NonNegativeReals,
    Suffix,
    sum_product,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


def c_rule(model, j):
    return 5 if j < 5 else 9.0 / 2


def b_rule(model, i):
    if i == 4:
        i = 5
    elif i == 5:
        i = 4
    return 5 if i < 5 else 7.0 / 2


def A_rule(model, i, j):
    if i == 4:
        i = 5
    elif i == 5:
        i = 4
    return 2 if i == j else 1


def primalcon_rule(model, i):
    return sum(model.A[i, j] * model.x[j] for j in model.N) >= model.b[i]


@register_model
class LP_unique_duals(_BaseTestModel):
    """
    A LP with unique dual values
    """

    description = "LP_unique_duals"
    capabilities = set(['linear'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = None
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        n = 7
        m = 7
        model.N = RangeSet(1, n)
        model.M = RangeSet(1, m)

        model.c = Param(model.N, rule=c_rule)

        model.b = Param(model.M, rule=b_rule)

        model.A = Param(model.M, model.N, rule=A_rule)

        model.x = Var(model.N, within=NonNegativeReals)
        model.y = Var(model.M, within=NonNegativeReals)

        model.cost = Objective(expr=sum_product(model.c, model.x))

        model.primalcon = Constraint(model.M, rule=primalcon_rule)

        # model.dual = Suffix(direction=Suffix.IMPORT)
        # model.rc = Suffix(direction=Suffix.IMPORT)
        model.slack = Suffix(direction=Suffix.IMPORT)
        model.urc = Suffix(direction=Suffix.IMPORT)
        model.lrc = Suffix(direction=Suffix.IMPORT)

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        for i in model.x:
            model.x[i] = None
        for i in model.y:
            model.y[i] = None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.core.base.matrix_constraint import MatrixConstraint


def _create_variable_list(size, **kwds):
    assert size > 0
    return pyo.Var(pyo.RangeSet(0, size - 1), **kwds)


def _get_csr(m, n, value):
    data = [value] * (m * n)
    indices = [j for j in range(n) for i in range(m)]
    indptr = [0]
    for i in range(m):
        indptr.append(indptr[-1] + n)
    return data, indices, indptr


class TestMatrixConstraint(unittest.TestCase):
    def test_init(self):
        m = pyo.ConcreteModel()
        m.v = _create_variable_list(3, initialize=1.0)
        data, indices, indptr = _get_csr(3, 3, 0.0)
        lb = [None] * 3
        ub = [None] * 3
        m.c = MatrixConstraint(data, indices, indptr, lb, ub, x=list(m.v.values()))
        self.assertEqual(len(m.c), 3)
        for k, c in m.c.items():
            with self.assertRaises(NotImplementedError):
                c.set_value(m.v[k] == 1)
            self.assertEqual(c.index(), k)
            self.assertEqual(c.strict_upper, False)
            self.assertEqual(c.strict_lower, False)
            self.assertEqual(c.lower, None)
            self.assertEqual(c.upper, None)
            self.assertEqual(c.equality, False)
            self.assertEqual(c.body(), 0)
            self.assertEqual(c(), 0)
            self.assertEqual(c.slack(), float('inf'))
            self.assertEqual(c.lslack(), float('inf'))
            self.assertEqual(c.uslack(), float('inf'))
            self.assertEqual(c.has_lb(), False)
            self.assertEqual(c.has_ub(), False)

        m = pyo.ConcreteModel()
        m.v = _create_variable_list(3, initialize=3)
        data, indices, indptr = _get_csr(2, 3, 1.0)
        m.c = MatrixConstraint(
            data, indices, indptr, lb=[0] * 2, ub=[2] * 2, x=list(m.v.values())
        )
        self.assertEqual(len(m.c), 2)
        for k, c in m.c.items():
            with self.assertRaises(NotImplementedError):
                c.set_value(m.v[k] == 1)
            self.assertEqual(c.index(), k)
            self.assertEqual(c.strict_upper, False)
            self.assertEqual(c.strict_lower, False)
            self.assertEqual(c.lower, 0)
            self.assertEqual(c.body(), 9)
            self.assertEqual(c(), 9)
            self.assertEqual(c.upper, 2)
            self.assertEqual(c.equality, False)

        m = pyo.ConcreteModel()
        m.v = _create_variable_list(3, initialize=3)
        data, indices, indptr = _get_csr(2, 3, 1.0)
        m.c = MatrixConstraint(
            data, indices, indptr, lb=[1] * 2, ub=[1] * 2, x=list(m.v.values())
        )
        self.assertEqual(len(m.c), 2)
        for k, c in m.c.items():
            with self.assertRaises(NotImplementedError):
                c.set_value(m.v[k] == 1)
            self.assertEqual(c.index(), k)
            self.assertEqual(c.strict_upper, False)
            self.assertEqual(c.strict_lower, False)
            self.assertEqual(c.lower, 1)
            self.assertEqual(c.body(), 9)
            self.assertEqual(c(), 9)
            self.assertEqual(c.upper, 1)
            self.assertEqual(c.equality, True)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from pyomo.contrib.gdpopt.util import get_main_elapsed_time
from pyomo.core import ConstraintList
from pyomo.opt import SolverFactory
from pyomo.contrib.mindtpy.config_options import _get_MindtPy_GOA_config
from pyomo.contrib.mindtpy.algorithm_base_class import _MindtPyAlgorithm
from pyomo.contrib.mindtpy.cut_generation import add_affine_cuts


@SolverFactory.register(
    'mindtpy.goa', doc='MindtPy: Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo'
)
class MindtPy_GOA_Solver(_MindtPyAlgorithm):
    """
    Decomposition solver for Mixed-Integer Nonlinear Programming (MINLP) problems.

    The MindtPy (Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo) solver
    applies a variety of decomposition-based approaches to solve Mixed-Integer
    Nonlinear Programming (MINLP) problems.
    This class includes:

    - Global outer approximation (GOA)
    - Global LP/NLP based branch-and-bound (GLP/NLP)
    """

    CONFIG = _get_MindtPy_GOA_config()

    def check_config(self):
        config = self.config
        config.add_slack = False
        config.use_mcpp = True
        config.equality_relaxation = False
        config.use_fbbt = True
        # add_no_good_cuts is True by default in GOA
        if not config.add_no_good_cuts and not config.use_tabu_list:
            config.add_no_good_cuts = True
            config.use_tabu_list = False
        # Set default initialization_strategy
        if config.single_tree:
            config.logger.info('Single-tree implementation is activated.')
            config.iteration_limit = 1
            config.add_slack = False
            if config.mip_solver not in {'cplex_persistent', 'gurobi_persistent'}:
                raise ValueError(
                    "Only cplex_persistent and gurobi_persistent are supported for LP/NLP based Branch and Bound method."
                    "Please refer to https://pyomo.readthedocs.io/en/stable/explanation/solvers/mindtpy.html#lp-nlp-based-branch-and-bound."
                )
            if config.threads > 1:
                config.threads = 1
                config.logger.info(
                    'The threads parameter is corrected to 1 since lazy constraint callback conflicts with multi-threads mode.'
                )

        super().check_config()

    def initialize_mip_problem(self):
        """Deactivate the nonlinear constraints to create the MIP problem."""
        super().initialize_mip_problem()
        self.mip.MindtPy_utils.cuts.aff_cuts = ConstraintList(doc='Affine cuts')

    def update_primal_bound(self, bound_value):
        """Update the primal bound.

        Call after solve fixed NLP subproblem.
        Use the optimal primal bound of the relaxed problem to update the dual bound.

        Parameters
        ----------
        bound_value : float
            The input value used to update the primal bound.
        """
        super().update_primal_bound(bound_value)
        self.primal_bound_progress_time.append(get_main_elapsed_time(self.timing))
        if self.primal_bound_improved:
            self.num_no_good_cuts_added.update(
                {self.primal_bound: len(self.mip.MindtPy_utils.cuts.no_good_cuts)}
            )

    def add_cuts(
        self,
        dual_values=None,
        linearize_active=True,
        linearize_violated=True,
        cb_opt=None,
        nlp=None,
    ):
        add_affine_cuts(self.mip, self.config, self.timing)

    def deactivate_no_good_cuts_when_fixing_bound(self, no_good_cuts):
        try:
            valid_no_good_cuts_num = self.num_no_good_cuts_added[self.primal_bound]
            if self.config.add_no_good_cuts:
                for i in range(valid_no_good_cuts_num + 1, len(no_good_cuts) + 1):
                    no_good_cuts[i].deactivate()
            if self.config.use_tabu_list:
                self.integer_list = self.integer_list[:valid_no_good_cuts_num]
        except KeyError as e:
            self.config.logger.error(e, exc_info=True)
            self.config.logger.error('Deactivating no-good cuts failed.')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to reformulate nonlinear models with linearity induced from
discrete variables.

Ref: Grossmann, IE; Voudouris, VT; Ghattas, O. Mixed integer linear
reformulations for some nonlinear discrete design optimization problems.

"""

import logging
import textwrap
from math import fabs

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    NonNegativeFloat,
    document_kwargs_from_configdict,
)
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.preprocessing.util import SuppressConstantObjectiveWarning
from pyomo.core import (
    Binary,
    Block,
    Constraint,
    Objective,
    Set,
    TransformationFactory,
    Var,
    summation,
    value,
)
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.gdp import Disjunct, Disjunction
from pyomo.opt import TerminationCondition as tc
from pyomo.opt import SolverFactory
from pyomo.repn import generate_standard_repn

logger = logging.getLogger('pyomo.contrib.preprocessing')


@TransformationFactory.register(
    'contrib.induced_linearity',
    doc="Reformulate nonlinear constraints with induced linearity.",
)
@document_kwargs_from_configdict('CONFIG')
class InducedLinearity(IsomorphicTransformation):
    """Reformulate nonlinear constraints with induced linearity.

    Finds continuous variables :math:`v` where :math:`v = d_1 + d_2 + d_3`,
    where :math:`d`'s are discrete variables. These continuous variables may
    participate nonlinearly in other expressions, which may then be induced to
    be linear.

    The overall algorithm flow can be summarized as:

    1. Detect effectively discrete variables and the constraints that
       imply discreteness.
    2. Determine the set of valid values for each effectively discrete variable
    3. Find nonlinear expressions in which effectively discrete variables
       participate.
    4. Reformulate nonlinear expressions appropriately.

    .. note:: Tasks 1 & 2 must incorporate scoping considerations (Disjuncts)

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock("contrib.induced_linearity")
    CONFIG.declare(
        'equality_tolerance',
        ConfigValue(
            default=1e-6,
            domain=NonNegativeFloat,
            description="Tolerance on equality constraints.",
        ),
    )
    CONFIG.declare(
        'pruning_solver',
        ConfigValue(
            default='glpk', description="Solver to use when pruning possible values."
        ),
    )

    def _apply_to(self, model, **kwds):
        """Apply the transformation to the given model."""
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)
        _process_subcontainers(model, config)
        _process_container(model, config)


def _process_subcontainers(blk, config):
    for disj in blk.component_data_objects(Disjunct, active=True, descend_into=False):
        _process_subcontainers(disj, config)
        _process_container(disj, config)


def _process_container(blk, config):
    if not hasattr(blk, '_induced_linearity_info'):
        blk._induced_linearity_info = Block()
    else:
        assert blk._induced_linearity_info.ctype == Block
    eff_discr_vars = detect_effectively_discrete_vars(blk, config.equality_tolerance)
    # TODO will need to go through this for each disjunct, since it does
    # not (should not) descend into Disjuncts.

    # Determine the valid values for the effectively discrete variables
    possible_var_values = determine_valid_values(blk, eff_discr_vars, config)

    # Collect find bilinear expressions that can be reformulated using
    # knowledge of effectively discrete variables
    bilinear_map = _bilinear_expressions(blk)

    # Relevant constraints are those with bilinear terms that involve
    # effectively_discrete_vars
    processed_pairs = ComponentSet()
    for v1, var_values in possible_var_values.items():
        v1_pairs = bilinear_map.get(v1, ())
        for v2, bilinear_constrs in v1_pairs.items():
            if (v1, v2) in processed_pairs:
                continue
            _process_bilinear_constraints(blk, v1, v2, var_values, bilinear_constrs)
            processed_pairs.add((v2, v1))
            # processed_pairs.add((v1, v2))  # TODO is this necessary?


def determine_valid_values(block, discr_var_to_constrs_map, config):
    """Calculate valid values for each effectively discrete variable.

    We need the set of possible values for the effectively discrete variable in
    order to do the reformulations.

    Right now, we select a naive approach where we look for variables in the
    discreteness-inducing constraints. We then adjust their values and see if
    things are still feasible. Based on their coefficient values, we can infer a
    set of allowable values for the effectively discrete variable.

    Args:
        block: The model or a disjunct on the model.

    """
    possible_values = ComponentMap()

    for eff_discr_var, constrs in discr_var_to_constrs_map.items():
        # get the superset of possible values by looking through the
        # constraints
        for constr in constrs:
            repn = generate_standard_repn(constr.body)
            var_coef = sum(
                coef
                for i, coef in enumerate(repn.linear_coefs)
                if repn.linear_vars[i] is eff_discr_var
            )
            const = -(repn.constant - constr.upper) / var_coef
            possible_vals = set((const,))
            for i, var in enumerate(repn.linear_vars):
                if var is eff_discr_var:
                    continue
                coef = -repn.linear_coefs[i] / var_coef
                if var.is_binary():
                    var_values = (0, coef)
                elif var.is_integer():
                    var_values = [v * coef for v in range(var.lb, var.ub + 1)]
                else:
                    raise ValueError(
                        '%s has unacceptable variable domain: %s'
                        % (var.name, var.domain)
                    )
                possible_vals = set(
                    (v1 + v2 for v1 in possible_vals for v2 in var_values)
                )
            old_possible_vals = possible_values.get(eff_discr_var, None)
            if old_possible_vals is not None:
                possible_values[eff_discr_var] = old_possible_vals & possible_vals
            else:
                possible_values[eff_discr_var] = possible_vals

    possible_values = prune_possible_values(block, possible_values, config)

    return possible_values


def prune_possible_values(block_scope, possible_values, config):
    # Prune the set of possible values by solving a series of feasibility
    # problems
    top_level_scope = block_scope.model()
    tmp_name = unique_component_name(top_level_scope, '_induced_linearity_prune_data')
    tmp_orig_blk = Block()
    setattr(top_level_scope, tmp_name, tmp_orig_blk)
    tmp_orig_blk._possible_values = possible_values
    tmp_orig_blk._possible_value_vars = list(v for v in possible_values)
    tmp_orig_blk._tmp_block_scope = (block_scope,)
    model = top_level_scope.clone()
    tmp_clone_blk = getattr(model, tmp_name)
    for obj in model.component_data_objects(Objective, active=True):
        obj.deactivate()
    for constr in model.component_data_objects(
        Constraint, active=True, descend_into=(Block, Disjunct)
    ):
        if constr.body.polynomial_degree() not in (1, 0):
            constr.deactivate()
    if block_scope.ctype == Disjunct:
        disj = tmp_clone_blk._tmp_block_scope[0]
        disj.indicator_var.fix(1)
        TransformationFactory('gdp.bigm').apply_to(model)
    # FIXME: this whole transformation should be reworked to solve
    # feasibility checks on independent models (using References) and
    # NOT have to clone the model / deactivate components in place.
    #
    # Deactivate any Disjuncts / Disjunctions (so the writers don't complain)
    for d in model.component_data_objects(Disjunction):
        d.deactivate()
    for d in model.component_data_objects(Disjunct):
        d._deactivate_without_fixing_indicator()
    tmp_clone_blk.test_feasible = Constraint()
    tmp_clone_blk._obj = Objective(expr=1)
    for eff_discr_var, vals in tmp_clone_blk._possible_values.items():
        val_feasible = {}
        for val in vals:
            tmp_clone_blk.test_feasible.set_value(eff_discr_var == val)
            with SuppressConstantObjectiveWarning():
                res = SolverFactory(config.pruning_solver).solve(model)
            if res.solver.termination_condition is tc.infeasible:
                val_feasible[val] = False
        tmp_clone_blk._possible_values[eff_discr_var] = set(
            v
            for v in tmp_clone_blk._possible_values[eff_discr_var]
            if val_feasible.get(v, True)
        )
    for i, var in enumerate(tmp_orig_blk._possible_value_vars):
        possible_values[var] = tmp_clone_blk._possible_values[
            tmp_clone_blk._possible_value_vars[i]
        ]

    return possible_values


def _process_bilinear_constraints(block, v1, v2, var_values, bilinear_constrs):
    # TODO check that the appropriate variable bounds exist.
    if not (v2.has_lb() and v2.has_ub()):
        logger.warning(textwrap.dedent("""\
            Attempting to transform bilinear term {v1} * {v2} using effectively
            discrete variable {v1}, but {v2} is missing a lower or upper bound:
            ({v2lb}, {v2ub}).
            """.format(v1=v1, v2=v2, v2lb=v2.lb, v2ub=v2.ub)).strip())
        return False
    blk = Block()
    unique_name = unique_component_name(
        block,
        ("%s_%s_bilinear" % (v1.local_name, v2.local_name))
        .replace('[', '')
        .replace(']', ''),
    )
    block._induced_linearity_info.add_component(unique_name, blk)
    # TODO think about not using floats as indices in a set
    blk.valid_values = Set(initialize=sorted(var_values))
    blk.x_active = Var(blk.valid_values, domain=Binary, initialize=1)
    blk.v_increment = Var(
        blk.valid_values, domain=v2.domain, bounds=(v2.lb, v2.ub), initialize=v2.value
    )
    blk.v_defn = Constraint(expr=v2 == summation(blk.v_increment))

    @blk.Constraint(blk.valid_values)
    def v_lb(blk, val):
        return v2.lb * blk.x_active[val] <= blk.v_increment[val]

    @blk.Constraint(blk.valid_values)
    def v_ub(blk, val):
        return blk.v_increment[val] <= v2.ub * blk.x_active[val]

    blk.select_one_value = Constraint(expr=summation(blk.x_active) == 1)
    # Categorize as case 1 or case 2
    for bilinear_constr in bilinear_constrs:
        # repn = generate_standard_repn(bilinear_constr.body)

        # Case 1: no other variables besides bilinear term in constraint. v1
        # (effectively discrete variable) is positive.
        # if (len(repn.quadratic_vars) == 1 and len(repn.linear_vars) == 0
        #         and repn.nonlinear_expr is None):
        #     _reformulate_case_1(v1, v2, discrete_constr, bilinear_constr)

        # NOTE: Case 1 is left unimplemented for now, because it involves some
        # messier logic with respect to how the transformation needs to happen.

        # Case 2: this is everything else, but do we want to have a special
        # case if there are nonlinear expressions involved with the constraint?
        pass
        _reformulate_case_2(blk, v1, v2, bilinear_constr)
    pass


def _reformulate_case_2(blk, v1, v2, bilinear_constr):
    repn = generate_standard_repn(bilinear_constr.body)
    replace_index = next(
        i
        for i, var_tup in enumerate(repn.quadratic_vars)
        if (var_tup[0] is v1 and var_tup[1] is v2)
        or (var_tup[0] is v2 and var_tup[1] is v1)
    )
    bilinear_constr.set_value(
        (
            bilinear_constr.lower,
            sum(coef * repn.linear_vars[i] for i, coef in enumerate(repn.linear_coefs))
            + repn.quadratic_coefs[replace_index]
            * sum(val * blk.v_increment[val] for val in blk.valid_values)
            + sum(
                repn.quadratic_coefs[i] * var_tup[0] * var_tup[1]
                for i, var_tup in enumerate(repn.quadratic_vars)
                if not i == replace_index
            )
            + repn.constant
            + zero_if_None(repn.nonlinear_expr),
            bilinear_constr.upper,
        )
    )


def zero_if_None(val):
    return 0 if val is None else val


def _bilinear_expressions(model):
    # TODO for now, we look for only expressions where the bilinearities are
    # exposed on the root level SumExpression, and thus accessible via
    # generate_standard_repn. This will not detect exp(x*y). We require a
    # factorization transformation to be applied beforehand in order to pick
    # these constraints up.
    pass
    # Bilinear map will be stored in the format:
    # x --> (y --> [constr1, constr2, ...], z --> [constr2, constr3])
    bilinear_map = ComponentMap()
    for constr in model.component_data_objects(
        Constraint, active=True, descend_into=(Block, Disjunct)
    ):
        if constr.body.polynomial_degree() in (1, 0):
            continue  # Skip trivial and linear constraints
        repn = generate_standard_repn(constr.body)
        for pair in repn.quadratic_vars:
            v1, v2 = pair
            v1_pairs = bilinear_map.get(v1, ComponentMap())
            if v2 in v1_pairs:
                # bilinear term has been found before. Simply add constraint to
                # the set associated with the bilinear term.
                v1_pairs[v2].add(constr)
            else:
                # We encounter the bilinear term for the first time.
                bilinear_map[v1] = v1_pairs
                bilinear_map[v2] = bilinear_map.get(v2, ComponentMap())
                constraints_with_bilinear_pair = ComponentSet([constr])
                bilinear_map[v1][v2] = constraints_with_bilinear_pair
                bilinear_map[v2][v1] = constraints_with_bilinear_pair
    return bilinear_map


def detect_effectively_discrete_vars(block, equality_tolerance):
    """Detect effectively discrete variables.

    These continuous variables are the sum of discrete variables.

    """
    # Map of effectively_discrete var --> inducing constraints
    effectively_discrete = ComponentMap()

    for constr in block.component_data_objects(Constraint, active=True):
        if constr.lower is None or constr.upper is None:
            continue  # skip inequality constraints
        if fabs(value(constr.lower) - value(constr.upper)) > equality_tolerance:
            continue  # not equality constraint. Skip.
        if constr.body.polynomial_degree() not in (1, 0):
            continue  # skip nonlinear expressions
        repn = generate_standard_repn(constr.body)
        if len(repn.linear_vars) < 2:
            # TODO should this be < 2 or < 1?
            # TODO we should make sure that trivial equality relations are
            # preprocessed before this, or we will end up reformulating
            # expressions that we do not need to here.
            continue
        non_discrete_vars = list(v for v in repn.linear_vars if v.is_continuous())
        if len(non_discrete_vars) == 1:
            # We know that this is an effectively discrete continuous
            # variable. Add it to our identified variable list.
            var = non_discrete_vars[0]
            inducing_constraints = effectively_discrete.get(var, [])
            inducing_constraints.append(constr)
            effectively_discrete[var] = inducing_constraints
        # TODO we should eventually also look at cases where all other
        # non_discrete_vars are effectively_discrete_vars

    return effectively_discrete



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import logging
import pyomo.common.unittest as unittest
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import (
    Any,
    Binary,
    Block,
    ConcreteModel,
    Constraint,
    Objective,
    Set,
    TransformationFactory,
    Var,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.util import _parent_disjunct
from pyomo.common.log import LoggingIntercept


class TestBoundPretransformation(unittest.TestCase):
    def create_nested_structure(self):
        """
        Creates a two-term Disjunction with on nested two-term Disjunction on
        the first Disjunct
        """
        m = ConcreteModel()
        m.outer_d1 = Disjunct()
        m.outer_d1.inner_d1 = Disjunct()
        m.outer_d1.inner_d2 = Disjunct()
        m.outer_d1.inner = Disjunction(expr=[m.outer_d1.inner_d1, m.outer_d1.inner_d2])
        m.outer_d2 = Disjunct()
        m.outer = Disjunction(expr=[m.outer_d1, m.outer_d2])
        return m

    def create_nested_model(self):
        """
        -100 <= x <= 102
        [-10 <= x <= 11, [x <= 3] v [x >= -7]] v [x == 0]
        """
        m = self.create_nested_structure()
        m.x = Var(bounds=(-100, 102))
        m.outer_d1.c = Constraint(expr=(-10, m.x, 11))
        m.outer_d1.inner_d1.c = Constraint(expr=m.x <= 3)
        m.outer_d1.inner_d2.c = Constraint(expr=m.x >= -7)
        m.outer_d2.c = Constraint(expr=m.x == 0)

        return m

    def check_nested_model_disjunction(self, m, bt):
        # We expect: -10w_1 -7w_2 <= x <= 3w_1 + 11w_2

        cons = bt.get_transformed_constraints(m.x, m.outer)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            -10.0 * m.outer_d1.binary_indicator_var
            + 0.0 * m.outer_d2.binary_indicator_var
            <= m.x,
        )
        assertExpressionsEqual(
            self,
            ub.expr,
            11.0 * m.outer_d1.binary_indicator_var
            + 0.0 * m.outer_d2.binary_indicator_var
            >= m.x,
        )

        cons = bt.get_transformed_constraints(m.x, m.outer_d1.inner)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            -10.0 * m.outer_d1.inner_d1.binary_indicator_var
            - 7.0 * m.outer_d1.inner_d2.binary_indicator_var
            <= m.x,
        )
        assertExpressionsEqual(
            self,
            ub.expr,
            3.0 * m.outer_d1.inner_d1.binary_indicator_var
            + 11.0 * m.outer_d1.inner_d2.binary_indicator_var
            >= m.x,
        )

        # All the disjunctive constraints were transformed
        self.assertFalse(m.outer_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d2.c.active)
        self.assertFalse(m.outer_d2.c.active)

    def test_transform_nested_model(self):
        m = self.create_nested_model()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)
        self.check_nested_model_disjunction(m, bt)

        # There aren't any other constraints on the model other than what we
        # added
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            4,
        )

    def test_transform_nested_model_no_0_terms(self):
        m = self.create_nested_model()
        m.outer_d2.c.deactivate()
        m.outer_d2.c2 = Constraint(expr=m.x == 101)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        # We expect: -10w_1 -7w_2 + 101 y_2 <= x <= 3w_1 + 11w_2 + 101y_2

        cons = bt.get_transformed_constraints(m.x, m.outer)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            -10.0 * m.outer_d1.binary_indicator_var
            + 101.0 * m.outer_d2.binary_indicator_var
            <= m.x,
        )
        assertExpressionsEqual(
            self,
            ub.expr,
            11.0 * m.outer_d1.binary_indicator_var
            + 101.0 * m.outer_d2.binary_indicator_var
            >= m.x,
        )

        cons = bt.get_transformed_constraints(m.x, m.outer_d1.inner)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            -10.0 * m.outer_d1.inner_d1.binary_indicator_var
            - 7.0 * m.outer_d1.inner_d2.binary_indicator_var
            <= m.x,
        )
        assertExpressionsEqual(
            self,
            ub.expr,
            3.0 * m.outer_d1.inner_d1.binary_indicator_var
            + 11.0 * m.outer_d1.inner_d2.binary_indicator_var
            >= m.x,
        )
        self.assertIs(_parent_disjunct(lb), m.outer_d1)
        self.assertIs(_parent_disjunct(ub), m.outer_d1)

        # All the disjunctive constraints were transformed
        self.assertFalse(m.outer_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d2.c.active)
        self.assertFalse(m.outer_d2.c2.active)

        # There aren't any other constraints on the model other than what we
        # added
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            4,
        )

    def test_transformation_gives_up_without_enough_bound_info(self):
        """
        If we have unbounded variables and not enough bounding constraints,
        we want the transformation to just leave the bounding constraints
        be to be transformed later.
        """
        m = self.create_nested_structure()
        m.x = Var()
        m.y = Var(bounds=(4, 67))
        m.outer_d1.c = Constraint(Any)
        m.outer_d1.c[1] = m.x >= 3
        m.outer_d1.c[2] = 5 <= m.y
        m.outer_d1.inner_d1.c = Constraint(Any)
        m.outer_d1.inner_d1.c[1] = m.x >= 4
        m.outer_d1.inner_d2.c = Constraint(Any)
        m.outer_d1.inner_d2.c[1] = m.x >= 17
        m.outer_d2.c = Constraint(Any)
        m.outer_d2.c[1] = m.x <= 1
        m.outer_d2.c[2] = m.y <= 66
        m.outer_d2.c[3] = m.x >= 2

        # y constraints should be fully transformed, and x can do lower but not
        # upper.

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        # We expect: 3d_1 + 2d_2 <= x
        #            5d_1 + 4d_2 <= y <= 67d_1 + 66d_2
        # in the global scope, and
        # x >= 4a_1 + 17a_2 on outer_d1

        cons = bt.get_transformed_constraints(m.x, m.outer)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            3.0 * m.outer_d1.binary_indicator_var
            + 2.0 * m.outer_d2.binary_indicator_var
            <= m.x,
        )

        cons = bt.get_transformed_constraints(m.y, m.outer)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            5.0 * m.outer_d1.binary_indicator_var + 4 * m.outer_d2.binary_indicator_var
            <= m.y,
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            67 * m.outer_d1.binary_indicator_var
            + 66.0 * m.outer_d2.binary_indicator_var
            >= m.y,
        )

        cons = bt.get_transformed_constraints(m.x, m.outer_d1.inner)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            4.0 * m.outer_d1.inner_d1.binary_indicator_var
            + 17.0 * m.outer_d1.inner_d2.binary_indicator_var
            <= m.x,
        )
        self.assertIs(_parent_disjunct(cons[0]), m.outer_d1)

        # check that all the y constraints are deactivated, and that the
        # lower bound ones for x are, but not the upper bound ones
        self.assertFalse(m.outer_d1.c[1].active)
        self.assertFalse(m.outer_d1.c[2].active)
        self.assertFalse(m.outer_d1.inner_d1.c[1].active)
        self.assertFalse(m.outer_d1.inner_d2.c[1].active)
        self.assertTrue(m.outer_d2.c[1].active)
        self.assertFalse(m.outer_d2.c[2].active)
        self.assertFalse(m.outer_d2.c[3].active)

        # and check that there are only four active constraints, the ones we
        # made and the remaining upper bound for x:
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            5,
        )

    def test_partially_deactivating_constraint_lb_transformed_but_ub_not(self):
        m = ConcreteModel()
        m.x = Var()
        m.x.setlb(3.0)
        m.d1 = Disjunct()
        m.d1.c = Constraint(expr=(4.5, m.x, 6))
        m.d2 = Disjunct()
        m.d2.c = Constraint(expr=5.6 <= m.x)
        m.disj = Disjunction(expr=[m.d1, m.d2])

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        # We expect: 4.5*y_1 + 5.6*y_2 <= x

        cons = bt.get_transformed_constraints(m.x, m.disj)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            4.5 * m.d1.binary_indicator_var + 5.6 * m.d2.binary_indicator_var <= m.x,
        )

        self.assertFalse(m.d1.c.active)
        self.assertFalse(m.d2.c.active)
        c_ub = m.d1.component('c_ub')
        self.assertIsInstance(c_ub, Constraint)
        self.assertTrue(c_ub.active)
        assertExpressionsEqual(self, c_ub.expr, m.x <= 6.0)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            2,
        )

    def test_partially_deactivating_constraint_ub_transformed_but_lb_not(self):
        m = ConcreteModel()
        m.w = Var()
        m.d = Disjunct([1, 2, 3])
        m.disjunction = Disjunction(expr=[m.d[1], m.d[2], m.d[3]])

        m.d[1].c = Constraint(expr=m.w == 45)
        m.d[2].c = Constraint(expr=m.w <= 36)
        m.d[3].c = Constraint(expr=m.w <= 232)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        # We expect: 45*y_1 + 36y_2 + 232y_3 >= w

        cons = bt.get_transformed_constraints(m.w, m.disjunction)
        self.assertEqual(len(cons), 1)
        ub = cons[0]
        assertExpressionsEqual(
            self,
            ub.expr,
            45.0 * m.d[1].binary_indicator_var
            + 36.0 * m.d[2].binary_indicator_var
            + 232.0 * m.d[3].binary_indicator_var
            >= m.w,
        )

        self.assertFalse(m.d[1].c.active)
        self.assertFalse(m.d[2].c.active)
        self.assertFalse(m.d[3].c.active)
        c_lb = m.d[1].component('c_lb')
        self.assertIsInstance(c_lb, Constraint)
        self.assertTrue(c_lb.active)
        assertExpressionsEqual(self, c_lb.expr, m.w >= 45.0)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            2,
        )

    def create_two_disjunction_model(self):
        m = self.create_nested_model()
        m.y = Var()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d3 = Disjunct()
        m.disjunction = Disjunction(expr=[m.d1, m.d2, m.d3])

        m.d1.c = Constraint(expr=m.y == 7.8)
        m.d1.c_x = Constraint(expr=m.x <= 27)
        m.d2.c = Constraint(expr=m.y == 8.9)
        m.d2.c_x = Constraint(expr=m.x >= 34)
        m.d3.c = Constraint(expr=m.y <= 45.7)
        return m

    def test_transform_multiple_disjunctions(self):
        m = self.create_two_disjunction_model()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        self.check_nested_model_disjunction(m, bt)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            -100 * m.d1.binary_indicator_var
            + 34.0 * m.d2.binary_indicator_var
            + -100 * m.d3.binary_indicator_var
            <= m.x,
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            27.0 * m.d1.binary_indicator_var
            + 102 * m.d2.binary_indicator_var
            + 102 * m.d3.binary_indicator_var
            >= m.x,
        )

        cons = bt.get_transformed_constraints(m.y, m.disjunction)
        self.assertEqual(len(cons), 1)
        ub = cons[0]
        assertExpressionsEqual(
            self,
            ub.expr,
            7.8 * m.d1.binary_indicator_var
            + 8.9 * m.d2.binary_indicator_var
            + 45.7 * m.d3.binary_indicator_var
            >= m.y,
        )

        self.assertFalse(m.d1.c.active)
        self.assertFalse(m.d1.c_x.active)
        self.assertFalse(m.d2.c.active)
        self.assertFalse(m.d2.c_x.active)
        self.assertFalse(m.d3.c.active)

        c_lb = m.d1.component('c_lb')
        self.assertIsInstance(c_lb, Constraint)
        self.assertTrue(c_lb.active)
        assertExpressionsEqual(self, c_lb.expr, 7.8 <= m.y)
        c_lb = m.d2.component('c_lb')
        self.assertIsInstance(c_lb, Constraint)
        self.assertTrue(c_lb.active)
        assertExpressionsEqual(self, c_lb.expr, 8.9 <= m.y)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            9,
        )

    def test_disjunction_target(self):
        m = self.create_two_disjunction_model()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m, targets=m.outer)

        self.check_nested_model_disjunction(m, bt)

        self.assertTrue(m.d1.c.active)
        self.assertTrue(m.d1.c_x.active)
        self.assertTrue(m.d2.c.active)
        self.assertTrue(m.d2.c_x.active)
        self.assertTrue(m.d3.c.active)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            9,
        )

    def test_get_transformed_constraint_errors(self):
        m = self.create_two_disjunction_model()
        m.z = Var()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m, targets=m.outer)

        out = StringIO()
        with LoggingIntercept(
            out, 'pyomo.gdp.plugins.bound_pretransformation', logging.DEBUG
        ):
            nothing = bt.get_transformed_constraints(m.z, m.outer)
        self.assertEqual(len(nothing), 0)
        # ...And we log that we're a bit confused.
        self.assertEqual(
            out.getvalue(),
            "Constraint bounding variable 'z' on Disjunction 'outer' was "
            "not transformed by the 'gdp.bound_pretransformation' "
            "transformation\n",
        )

        out = StringIO()
        with LoggingIntercept(
            out, 'pyomo.gdp.plugins.bound_pretransformation', logging.DEBUG
        ):
            nothing = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(nothing), 0)
        self.assertEqual(
            out.getvalue(),
            "No variable on Disjunction 'disjunction' was transformed with the "
            "gdp.bound_pretransformation transformation\n",
        )

    def test_univariate_constraints_with_expression_bodies(self):
        m = self.create_nested_structure()

        # This is a convoluted way to write the same model as the nested model
        m.x = Var(bounds=(-100, 102))
        m.outer_d1.c = Constraint(expr=-20 <= 2 * m.x)
        m.outer_d1.c2 = Constraint(expr=m.x - 1 <= 10)
        m.outer_d1.inner_d1.c = Constraint(expr=3 * m.x - 7 <= 2)
        m.outer_d1.inner_d2.c = Constraint(expr=m.x >= -7)
        m.outer_d2.c = Constraint(expr=m.x + 4 == 4)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        self.check_nested_model_disjunction(m, bt)

        self.assertFalse(m.outer_d1.c.active)
        self.assertFalse(m.outer_d1.c2.active)
        self.assertFalse(m.outer_d1.inner_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d2.c.active)
        self.assertFalse(m.outer_d2.c.active)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            4,
        )

    def test_bound_constraints_skip_levels_in_hierarchy(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 10))
        m.y = Var()
        m.Y = Disjunct([1, 2])
        m.Z = Disjunct([1, 2, 3])
        m.W = Disjunct([1, 2])
        m.W[1].c = Constraint(expr=m.x <= 7)
        m.W[2].c = Constraint(expr=m.x <= 9)
        m.Z[1].c = Constraint(expr=m.y == 0)
        m.Z[1].w_disj = Disjunction(expr=[m.W[i] for i in [1, 2]])
        m.Z[2].c = Constraint(expr=m.y == 1)
        m.Z[3].c = Constraint(expr=m.y == 2)
        m.Y[1].c = Constraint(expr=m.x >= 2)
        m.Y[1].z_disj = Disjunction(expr=[m.Z[i] for i in [1, 2, 3]])
        m.Y[2].c1 = Constraint(expr=m.x == 0)
        m.Y[2].c2 = Constraint(expr=(3, m.y, 17))
        m.y_disj = Disjunction(expr=[m.Y[i] for i in [1, 2]])

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.y_disj)
        self.assertEqual(len(cons), 2)
        x_lb = cons[0]
        assertExpressionsEqual(
            self,
            x_lb.expr,
            2.0 * m.Y[1].binary_indicator_var + 0 * m.Y[2].binary_indicator_var <= m.x,
        )
        x_ub = cons[1]
        assertExpressionsEqual(
            self,
            x_ub.expr,
            10 * m.Y[1].binary_indicator_var + 0.0 * m.Y[2].binary_indicator_var >= m.x,
        )
        self.assertIsNone(_parent_disjunct(x_lb))
        self.assertIsNone(_parent_disjunct(x_ub))

        cons = bt.get_transformed_constraints(m.y, m.y_disj)
        self.assertEqual(len(cons), 0)

        cons = bt.get_transformed_constraints(m.y, m.Y[1].z_disj)
        self.assertEqual(len(cons), 2)
        y_lb = cons[0]
        assertExpressionsEqual(
            self,
            y_lb.expr,
            0.0 * m.Z[1].binary_indicator_var
            + 1.0 * m.Z[2].binary_indicator_var
            + 2.0 * m.Z[3].binary_indicator_var
            <= m.y,
        )
        y_ub = cons[1]
        assertExpressionsEqual(
            self,
            y_ub.expr,
            0.0 * m.Z[1].binary_indicator_var
            + 1.0 * m.Z[2].binary_indicator_var
            + 2.0 * m.Z[3].binary_indicator_var
            >= m.y,
        )

        cons = bt.get_transformed_constraints(m.x, m.Y[1].z_disj)
        self.assertEqual(len(cons), 0)

        cons = bt.get_transformed_constraints(m.y, m.Z[1].w_disj)
        self.assertEqual(len(cons), 0)

        cons = bt.get_transformed_constraints(m.x, m.Z[1].w_disj)
        self.assertEqual(len(cons), 2)
        x_lb = cons[0]
        assertExpressionsEqual(
            self,
            x_lb.expr,
            2.0 * m.W[1].binary_indicator_var + 2.0 * m.W[2].binary_indicator_var
            <= m.x,
        )
        x_ub = cons[1]
        assertExpressionsEqual(
            self,
            x_ub.expr,
            7.0 * m.W[1].binary_indicator_var + 9.0 * m.W[2].binary_indicator_var
            >= m.x,
        )

        self.assertFalse(m.W[1].c.active)
        self.assertFalse(m.W[2].c.active)
        self.assertFalse(m.Z[1].c.active)
        self.assertFalse(m.Z[2].c.active)
        self.assertFalse(m.Z[3].c.active)
        self.assertFalse(m.Y[1].c.active)
        self.assertFalse(m.Y[2].c1.active)
        self.assertTrue(m.Y[2].c2.active)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            7,
        )

    def test_skip_nonlinear_and_multivariate_constraints(self):
        m = self.create_nested_model()
        m.y = Var()
        m.z = Var()
        m.outer_d1.nonlinear = Constraint(expr=m.y**2 <= 7)
        m.outer_d1.inner_d2.multivariate = Constraint(expr=m.x + m.y <= m.z)
        m.outer_d2.leave_it = Constraint(expr=m.z == 7)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        self.check_nested_model_disjunction(m, bt)

        self.assertTrue(m.outer_d1.nonlinear.active)
        self.assertTrue(m.outer_d1.inner_d2.multivariate.active)
        self.assertTrue(m.outer_d2.leave_it.active)

        self.assertFalse(m.outer_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d1.c.active)
        self.assertFalse(m.outer_d1.inner_d2.c.active)
        self.assertFalse(m.outer_d2.c.active)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            7,
        )

    def test_tightest_bound_is_at_root(self):
        """
        x >= 60
        [ [x >= 55, [ ] v [x >= 66] ] ] v [x >= 5]
        """
        m = ConcreteModel()
        m.x = Var()
        m.x.setlb(4)
        m.c = Constraint(expr=m.x >= 60)
        m.d = Disjunct([1, 2])
        m.inner1 = Disjunct([1, 2])
        m.inner2 = Disjunct([1, 2])
        m.disjunction = Disjunction(expr=[m.d[1], m.d[2]])
        m.d[1].disjunction = Disjunction(expr=[m.inner1[1], m.inner1[2]])
        m.inner1[1].disjunction = Disjunction(expr=[m.inner2[1], m.inner2[2]])

        m.d[2].c = Constraint(expr=m.x >= 5)
        m.inner1[1].c = Constraint(expr=m.x >= 55)
        m.inner2[2].c = Constraint(expr=m.x >= 66)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        print(lb.expr)
        assertExpressionsEqual(
            self,
            lb.expr,
            60.0 * m.d[1].binary_indicator_var + 60.0 * m.d[2].binary_indicator_var
            <= m.x,
        )
        self.assertIsNone(_parent_disjunct(lb))

        cons = bt.get_transformed_constraints(m.x, m.d[1].disjunction)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            60.0 * m.inner1[1].binary_indicator_var
            + 60.0 * m.inner1[2].binary_indicator_var
            <= m.x,
        )
        self.assertIs(_parent_disjunct(lb), m.d[1])

        cons = bt.get_transformed_constraints(m.x, m.inner1[1].disjunction)
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            60.0 * m.inner2[1].binary_indicator_var
            + 66.0 * m.inner2[2].binary_indicator_var
            <= m.x,
        )
        self.assertIs(_parent_disjunct(lb), m.inner1[1])

        # We shouldn't deactivate global constraints. Reason 1 being that we
        # don't deactivate bounds and Reason 2 being that generally the global
        # part of the model is none of pyomo.gdp's beeswax.
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            4,
        )

    def test_bounds_on_disjuncts_with_block_hierarchies(self):
        m = ConcreteModel()
        m.x = Var()
        m.b = Block()
        m.b.c = Constraint(expr=m.x <= 4)
        m.d = Disjunct([1, 2])
        m.d[1].b = Block()
        m.d[1].b.c = Constraint(expr=m.x <= 5)
        m.d[2].b = Block()
        m.d[2].b.c = Constraint(expr=m.x <= 3)
        m.d[2].c = Constraint(expr=m.x <= 4.1)
        m.disjunction = Disjunction(expr=[m.d[1], m.d[2]])

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 1)
        ub = cons[0]

        assertExpressionsEqual(
            self,
            ub.expr,
            4.0 * m.d[1].binary_indicator_var + 3.0 * m.d[2].binary_indicator_var
            >= m.x,
        )
        # just the one we made and the global one are active.
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            2,
        )

    def test_indexed_disjunction_target(self):
        m = ConcreteModel()
        m.x = Var()
        m.d = Disjunct([1, 2, 3, 4, 5])
        m.d[1].c = Constraint(expr=m.x <= 1)
        m.d[2].c = Constraint(expr=m.x <= 2)
        m.d[3].c = Constraint(expr=m.x <= 3)
        m.d[4].c = Constraint(expr=m.x >= -5)
        m.d[5].c = Constraint(expr=m.x >= -8)
        m.disjunction = Disjunction(['pos', 'neg'])
        m.disjunction['pos'] = [m.d[1], m.d[2], m.d[3]]
        m.disjunction['neg'] = [m.d[4], m.d[5]]

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m, targets=m.disjunction)

        cons = bt.get_transformed_constraints(m.x, m.disjunction['pos'])
        self.assertEqual(len(cons), 1)
        ub = cons[0]
        assertExpressionsEqual(
            self,
            ub.expr,
            1.0 * m.d[1].binary_indicator_var
            + 2.0 * m.d[2].binary_indicator_var
            + 3.0 * m.d[3].binary_indicator_var
            >= m.x,
        )
        cons = bt.get_transformed_constraints(m.x, m.disjunction['neg'])
        self.assertEqual(len(cons), 1)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            -5.0 * m.d[4].binary_indicator_var - 8.0 * m.d[5].binary_indicator_var
            <= m.x,
        )

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, descend_into=(Block, Disjunct), active=True
                    )
                )
            ),
            2,
        )

    def test_nested_target(self):
        m = self.create_nested_model()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m, targets=[m.outer_d1.inner])

        cons = bt.get_transformed_constraints(m.x, m.outer_d1.inner)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            -100 * m.outer_d1.inner_d1.binary_indicator_var
            - 7.0 * m.outer_d1.inner_d2.binary_indicator_var
            <= m.x,
        )
        self.assertIs(lb.parent_block().parent_block(), m.outer_d1)
        assertExpressionsEqual(
            self,
            ub.expr,
            3.0 * m.outer_d1.inner_d1.binary_indicator_var
            + 102 * m.outer_d1.inner_d2.binary_indicator_var
            >= m.x,
        )
        self.assertIs(ub.parent_block().parent_block(), m.outer_d1)

        self.assertTrue(m.outer_d1.c.active)
        self.assertTrue(m.outer_d2.c.active)
        self.assertTrue(lb.active)
        self.assertTrue(ub.active)

        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            4,
        )

    def test_targets_nested_in_each_other(self):
        m = self.create_nested_model()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m, targets=[m.outer_d1.inner, m.outer])

        # This should do the outermost disjunctions only--we should
        # get the same result as if we had transformed the whole
        # model.

        self.check_nested_model_disjunction(m, bt)
        # There aren't any other constraints on the model other than what we
        # added
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            4,
        )

    def test_variables_not_in_any_leaves(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.disjunct1 = Disjunct()
        m.disjunct1.c = Constraint(expr=m.x <= 9.7)
        m.disjunct1.disjunct1 = Disjunct()
        m.disjunct1.disjunct1.c = Constraint(expr=m.x + m.y <= 4)
        m.disjunct1.disjunct2 = Disjunct()
        m.disjunct1.disjunct2.c = Constraint(expr=m.y <= 9)
        m.disjunct1.disjunction = Disjunction(
            expr=[m.disjunct1.disjunct1, m.disjunct1.disjunct2]
        )
        m.disjunct2 = Disjunct()
        m.disjunct2.c = Constraint(expr=m.x <= 9)
        m.disjunction = Disjunction(expr=[m.disjunct1, m.disjunct2])

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 1)
        ub = cons[0]
        assertExpressionsEqual(
            self,
            ub.expr,
            9.7 * m.disjunct1.binary_indicator_var
            + 9.0 * m.disjunct2.binary_indicator_var
            >= m.x,
        )

        cons = bt.get_transformed_constraints(m.y, m.disjunction)
        self.assertEqual(len(cons), 0)

        self.assertFalse(m.disjunct1.c.active)
        self.assertFalse(m.disjunct2.c.active)

        # two leftover and the one we added
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            3,
        )

    def test_fixed_vars_handled_correctly(self):
        m = ConcreteModel()
        m.x = Var()
        m.x.setub(78)
        m.y = Var()
        m.y.fix(1)
        m.z = Var()

        m.disjunction = Disjunction(expr=[[m.x + m.y <= 5], [m.x <= 17], [m.z == 0]])

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 1)
        ub = cons[0]
        assertExpressionsEqual(
            self,
            ub.expr,
            4.0 * m.disjunction.disjuncts[0].binary_indicator_var
            + 17.0 * m.disjunction.disjuncts[1].binary_indicator_var
            + 78 * m.disjunction.disjuncts[2].binary_indicator_var
            >= m.x,
        )

        self.assertFalse(m.disjunction.disjuncts[0].constraint[1].active)
        self.assertFalse(m.disjunction.disjuncts[1].constraint[1].active)
        # Only have the one we added and the one on z.
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        Constraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            2,
        )

    def test_bounds_flipped_for_negative_coefs(self):
        m = ConcreteModel()
        m.Time = Set(initialize=range(4))
        m.makespan = Var(bounds=(0, 4))
        m.act_time = Var(m.Time, domain=Binary)
        m.cost = Var(bounds=(1, 6))

        @m.Disjunct()
        def d1(d):
            d.c = Constraint(expr=m.cost == 6)

        @m.Disjunct()
        def d2(d):
            d.c = Constraint(expr=m.cost == 5)
            d.do_act = Constraint(expr=sum(m.act_time[t] for t in m.Time) == 1)

            @d.Constraint(m.Time)
            def ms(d, t):
                return t * m.act_time[t] + 2 <= m.makespan

        m.disjunction = Disjunction(expr=[m.d1, m.d2])

        m.obj = Objective(expr=m.cost)

        m.act_time.fix(0)
        m.act_time[2].unfix()

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.act_time[2], m.disjunction)
        self.assertEqual(len(cons), 2)

        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            0 * m.disjunction.disjuncts[0].binary_indicator_var
            + m.disjunction.disjuncts[1].binary_indicator_var
            <= m.act_time[2],
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            m.disjunction.disjuncts[0].binary_indicator_var
            + m.disjunction.disjuncts[1].binary_indicator_var
            >= m.act_time[2],
        )

        cons = bt.get_transformed_constraints(m.act_time[0], m.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.act_time[1], m.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.act_time[3], m.disjunction)
        self.assertEqual(len(cons), 0)

        cons = bt.get_transformed_constraints(m.makespan, m.disjunction)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            0 * m.disjunction.disjuncts[0].binary_indicator_var
            + 2.0 * m.disjunction.disjuncts[1].binary_indicator_var
            <= m.makespan,
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            4 * m.disjunction.disjuncts[0].binary_indicator_var
            + 4 * m.disjunction.disjuncts[1].binary_indicator_var
            >= m.makespan,
        )

        cons = bt.get_transformed_constraints(m.cost, m.disjunction)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            6.0 * m.disjunction.disjuncts[0].binary_indicator_var
            + 5.0 * m.disjunction.disjuncts[1].binary_indicator_var
            <= m.cost,
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            6 * m.disjunction.disjuncts[0].binary_indicator_var
            + 5.0 * m.disjunction.disjuncts[1].binary_indicator_var
            >= m.cost,
        )

    def test_bounds_correct_for_multiple_nested_disjunctions(self):
        m = ConcreteModel()
        m.c = Var(bounds=(3, 9))
        m.x = Var(bounds=(0, 10))
        m.y = Var(bounds=(-10, 2))

        m.d1 = Disjunct()
        m.d1.cons = Constraint(expr=m.c == 4)
        m.d1.disjunction = Disjunction(expr=[[m.x + m.y >= 8], [m.x + m.y <= 3]])
        m.d1.disjunction2 = Disjunction(
            expr=[[m.x + 2 * m.y <= 4], [m.y + 2 * m.x >= 7]]
        )

        m.d2 = Disjunct()
        m.d2.cons = Constraint(expr=m.c == 5)
        m.d2.disjunction = Disjunction(expr=[[m.x + m.y >= 10], [m.x + m.y <= 0]])
        m.d2.disjunction2 = Disjunction(
            expr=[[m.x + 3 * m.y <= 2], [m.y + 2 * m.x >= 9]]
        )
        m.disjunction = Disjunction(expr=[m.d1, m.d2])

        # will be 4 if you solve
        m.obj = Objective(expr=m.c)

        bt = TransformationFactory('gdp.bound_pretransformation')
        bt.apply_to(m)

        cons = bt.get_transformed_constraints(m.x, m.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.y, m.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.c, m.disjunction)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        assertExpressionsEqual(
            self,
            lb.expr,
            4.0 * m.d1.binary_indicator_var + 5.0 * m.d2.binary_indicator_var <= m.c,
        )
        ub = cons[1]
        assertExpressionsEqual(
            self,
            ub.expr,
            4.0 * m.d1.binary_indicator_var + 5.0 * m.d2.binary_indicator_var >= m.c,
        )

        cons = bt.get_transformed_constraints(m.x, m.d1.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.y, m.d1.disjunction)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.c, m.d1.disjunction)
        self.assertEqual(len(cons), 0)

        cons = bt.get_transformed_constraints(m.x, m.d1.disjunction2)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.y, m.d1.disjunction2)
        self.assertEqual(len(cons), 0)
        cons = bt.get_transformed_constraints(m.c, m.d1.disjunction2)
        self.assertEqual(len(cons), 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import pyomo.common.unittest as unittest
from pyomo.common.dependencies import numpy as np, numpy_available
from pyomo.common.errors import InfeasibleConstraintException
import pyomo.contrib.fbbt.interval as interval

try:
    isfinite = math.isfinite
except AttributeError:
    # isfinite() was added to math in Python 3.2
    def isfinite(x):
        return not (math.isnan(x) or math.isinf(x))


class IntervalTestBase:
    """
    These tests are set up weird, but it is for a good reason.
    The interval arithmetic code is duplicated in pyomo.contrib.appsi for
    improved performance. We want to keep this version because
    it does not require building an extension. However, when we
    fix a bug in one module, we want to ensure we fix that bug
    in the other module. Therefore, we use this base class
    for testing both modules. The only difference in the
    derived classes is in the self.add, self.sub,
    self.mul, etc. attributes.
    """

    def setUp(self):
        if numpy_available:
            np.random.seed(0)

    def test_add(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        xl = -2.5
        xu = 2.8
        yl = -3.2
        yu = 2.7
        zl, zu = self.add(xl, xu, yl, yu)
        x = np.linspace(xl, xu, 100)
        y = np.linspace(yl, yu, 100)
        for _x in x:
            _z = _x + y
            self.assertTrue(np.all(zl <= _z))
            self.assertTrue(np.all(zu >= _z))

    def test_sub(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        xl = -2.5
        xu = 2.8
        yl = -3.2
        yu = 2.7
        zl, zu = self.sub(xl, xu, yl, yu)
        x = np.linspace(xl, xu, 100)
        y = np.linspace(yl, yu, 100)
        for _x in x:
            _z = _x - y
            self.assertTrue(np.all(zl <= _z))
            self.assertTrue(np.all(zu >= _z))

    def test_mul(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        xl = -2.5
        xu = 2.8
        yl = -3.2
        yu = 2.7
        zl, zu = self.mul(xl, xu, yl, yu)
        x = np.linspace(xl, xu, 100)
        y = np.linspace(yl, yu, 100)
        for _x in x:
            _z = _x * y
            self.assertTrue(np.all(zl <= _z))
            self.assertTrue(np.all(zu >= _z))

    def test_inv(self):
        lb, ub = self.inv(0.1, 0.2, 1e-8)
        self.assertAlmostEqual(lb, 5)
        self.assertAlmostEqual(ub, 10)

        lb, ub = self.inv(0, 0.1, 1e-8)
        self.assertEqual(lb, 10)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.inv(0, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.inv(-0.1, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, -10)

        lb, ub = self.inv(-0.2, -0.1, 1e-8)
        self.assertAlmostEqual(lb, -10)
        self.assertAlmostEqual(ub, -5)

        lb, ub = self.inv(0, -1e-16, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.inv(1e-16, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.inv(-1, 1, 1e-8)
        self.assertAlmostEqual(lb, -interval.inf)
        self.assertAlmostEqual(ub, interval.inf)

    def test_div(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        x_bounds = [(np.random.uniform(-5, -2), np.random.uniform(2, 5))]
        y_bounds = [
            (np.random.uniform(-5, -2), np.random.uniform(2, 5)),
            (0, np.random.uniform(2, 5)),
            (np.random.uniform(0, 2), np.random.uniform(2, 5)),
            (np.random.uniform(-5, -2), 0),
            (np.random.uniform(-5, -2), np.random.uniform(-2, 0)),
        ]
        for xl, xu in x_bounds:
            for yl, yu in y_bounds:
                zl, zu = self.div(xl, xu, yl, yu, 1e-8)
                x = np.linspace(xl, xu, 100)
                y = np.linspace(yl, yu, 100)
                for _x in x:
                    _z = _x / y
                    self.assertTrue(np.all(zl <= _z))
                    self.assertTrue(np.all(zu >= _z))

    def test_div_edge_cases(self):
        lb, ub = self.div(0, -1e-16, 0, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.div(0, 1e-16, 0, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.div(-1e-16, 0, 0, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

        lb, ub = self.div(1e-16, 0, 0, 0, 1e-8)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)

    def test_pow(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        x_bounds = [
            (np.random.uniform(0, 2), np.random.uniform(2, 5)),
            (0, np.random.uniform(2, 5)),
            (0, 0),
        ]
        y_bounds = [
            (np.random.uniform(-5, -2), np.random.uniform(2, 5)),
            (0, np.random.uniform(2, 5)),
            (np.random.uniform(0, 2), np.random.uniform(2, 5)),
            (np.random.uniform(-5, -2), 0),
            (np.random.uniform(-5, -2), np.random.uniform(-2, 0)),
        ]
        for xl, xu in x_bounds:
            for yl, yu in y_bounds:
                zl, zu = self.power(xl, xu, yl, yu, 1e-8)
                if xl == 0 and xu == 0 and yu < 0:
                    self.assertEqual(zl, -interval.inf)
                    self.assertEqual(zu, interval.inf)
                x = np.linspace(xl, xu, 100)
                y = np.linspace(yl, yu, 100)
                for _x in x:
                    _z = _x**y
                    self.assertTrue(np.all(zl - 1e-14 <= _z))
                    self.assertTrue(np.all(zu + 1e-14 >= _z))

        x_bounds = [
            (np.random.uniform(-5, -2), np.random.uniform(2, 5)),
            (np.random.uniform(-5, -2), np.random.uniform(-2, 0)),
            (np.random.uniform(-5, -2), 0),
        ]
        y_bounds = list(range(-4, 4))
        for xl, xu in x_bounds:
            for yl in y_bounds:
                yu = yl
                zl, zu = self.power(xl, xu, yl, yu, 1e-8)
                x = np.linspace(xl, xu, 100, endpoint=False)
                y = yl
                _z = x**y
                self.assertTrue(np.all(zl <= _z))
                self.assertTrue(np.all(zu >= _z))

    def test_pow2(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        xl = np.linspace(-2, 2, 9)
        xu = np.linspace(-2, 2, 9)
        yl = np.linspace(-2, 2, 9)
        yu = np.linspace(-2, 2, 9)
        for _xl in xl:
            for _xu in xu:
                if _xl > _xu:
                    continue
                for _yl in yl:
                    for _yu in yu:
                        if _yl > _yu:
                            continue
                        if _xl == 0 and _xu == 0 and _yu < 0:
                            lb, ub = self.power(_xl, _xu, _yl, _yu, 1e-8)
                            self.assertEqual(lb, -interval.inf)
                            self.assertEqual(ub, interval.inf)
                        elif (
                            _yl == _yu
                            and _yl != round(_yl)
                            and (_xu < 0 or (_xu < 0 and _yu < 0))
                        ):
                            with self.assertRaises(
                                (
                                    InfeasibleConstraintException,
                                    interval.IntervalException,
                                )
                            ):
                                lb, ub = self.power(_xl, _xu, _yl, _yu, 1e-8)
                        else:
                            lb, ub = self.power(_xl, _xu, _yl, _yu, 1e-8)
                            if isfinite(lb) and isfinite(ub):
                                nan_fill = 0.5 * (lb + ub)
                            elif isfinite(lb):
                                nan_fill = lb + 1
                            elif isfinite(ub):
                                nan_fill = ub - 1
                            else:
                                nan_fill = 0
                            x = np.linspace(_xl, _xu, 30)
                            y = np.linspace(_yl, _yu, 30)
                            z = x ** np.split(y, len(y))
                            z[np.isnan(z)] = nan_fill
                            all_values = z
                            estimated_lb = all_values.min()
                            estimated_ub = all_values.max()
                            self.assertTrue(lb - 1e-8 <= estimated_lb)
                            self.assertTrue(ub + 1e-8 >= estimated_ub)

    def test_exp(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        xl = -2.5
        xu = 2.8
        zl, zu = self.exp(xl, xu)
        x = np.linspace(xl, xu, 100)
        _z = np.exp(x)
        self.assertTrue(np.all(zl <= _z))
        self.assertTrue(np.all(zu >= _z))

    def test_log(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        x_bounds = [
            (np.random.uniform(0, 2), np.random.uniform(2, 5)),
            (0, np.random.uniform(2, 5)),
            (0, 0),
        ]
        for xl, xu in x_bounds:
            zl, zu = self.log(xl, xu)
            x = np.linspace(xl, xu, 100)
            _z = np.log(x)
            self.assertTrue(np.all(zl <= _z))
            self.assertTrue(np.all(zu >= _z))

    def test_cos(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        lbs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        ubs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        for xl in lbs:
            for xu in ubs:
                if xu >= xl:
                    zl, zu = self.cos(xl, xu)
                    x = np.linspace(xl, xu, 100)
                    _z = np.cos(x)
                    self.assertTrue(np.all(zl <= _z + 1e-14))
                    self.assertTrue(np.all(zu >= _z - 1e-14))

    def test_sin(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        lbs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        ubs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        for xl in lbs:
            for xu in ubs:
                if xu >= xl:
                    zl, zu = self.sin(xl, xu)
                    x = np.linspace(xl, xu, 100)
                    _z = np.sin(x)
                    self.assertTrue(np.all(zl <= _z + 1e-14))
                    self.assertTrue(np.all(zu >= _z - 1e-14))

    def test_tan(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        lbs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        ubs = np.linspace(-2 * math.pi, 2 * math.pi, 10)
        for xl in lbs:
            for xu in ubs:
                if xu >= xl:
                    zl, zu = self.tan(xl, xu)
                    x = np.linspace(xl, xu, 100)
                    _z = np.tan(x)
                    self.assertTrue(np.all(zl <= _z + 1e-14))
                    self.assertTrue(np.all(zu >= _z - 1e-14))

    def test_asin(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        yl, yu = self.asin(-0.5, 0.5, -interval.inf, interval.inf, 1e-8)
        self.assertEqual(yl, -interval.inf)
        self.assertEqual(yu, interval.inf)
        yl, yu = self.asin(-0.5, 0.5, -math.pi, math.pi, 1e-8)
        self.assertAlmostEqual(yl, -math.pi, 12)
        self.assertAlmostEqual(yu, math.pi, 12)
        yl, yu = self.asin(-0.5, 0.5, -math.pi / 2, math.pi / 2, 1e-8)
        self.assertAlmostEqual(yl, math.asin(-0.5))
        self.assertAlmostEqual(yu, math.asin(0.5))
        yl, yu = self.asin(-0.5, 0.5, -math.pi / 2 - 0.1, math.pi / 2 + 0.1, 1e-8)
        self.assertAlmostEqual(yl, math.asin(-0.5))
        self.assertAlmostEqual(yu, math.asin(0.5))
        yl, yu = self.asin(-0.5, 0.5, -math.pi / 2 + 0.1, math.pi / 2 - 0.1, 1e-8)
        self.assertAlmostEqual(yl, math.asin(-0.5))
        self.assertAlmostEqual(yu, math.asin(0.5))
        yl, yu = self.asin(-0.5, 0.5, -1.5 * math.pi, 1.5 * math.pi, 1e-8)
        self.assertAlmostEqual(yl, -3.6651914291880920, 12)
        self.assertAlmostEqual(yu, 3.6651914291880920, 12)
        yl, yu = self.asin(-0.5, 0.5, -1.5 * math.pi - 0.1, 1.5 * math.pi + 0.1, 1e-8)
        self.assertAlmostEqual(yl, -3.6651914291880920, 12)
        self.assertAlmostEqual(yu, 3.6651914291880920, 12)
        yl, yu = self.asin(-0.5, 0.5, -1.5 * math.pi + 0.1, 1.5 * math.pi - 0.1, 1e-8)
        self.assertAlmostEqual(yl, -3.6651914291880920, 12)
        self.assertAlmostEqual(yu, 3.6651914291880920, 12)

    def test_acos(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        yl, yu = self.acos(-0.5, 0.5, -interval.inf, interval.inf, 1e-8)
        self.assertEqual(yl, -interval.inf)
        self.assertEqual(yu, interval.inf)
        yl, yu = self.acos(-0.5, 0.5, -0.5 * math.pi, 0.5 * math.pi, 1e-8)
        self.assertAlmostEqual(yl, -0.5 * math.pi, 12)
        self.assertAlmostEqual(yu, 0.5 * math.pi, 12)
        yl, yu = self.acos(-0.5, 0.5, 0, math.pi, 1e-8)
        self.assertAlmostEqual(yl, math.acos(0.5))
        self.assertAlmostEqual(yu, math.acos(-0.5))
        yl, yu = self.acos(-0.5, 0.5, 0 - 0.1, math.pi + 0.1, 1e-8)
        self.assertAlmostEqual(yl, math.acos(0.5))
        self.assertAlmostEqual(yu, math.acos(-0.5))
        yl, yu = self.acos(-0.5, 0.5, 0 + 0.1, math.pi - 0.1, 1e-8)
        self.assertAlmostEqual(yl, math.acos(0.5))
        self.assertAlmostEqual(yu, math.acos(-0.5))
        yl, yu = self.acos(-0.5, 0.5, -math.pi, 0, 1e-8)
        self.assertAlmostEqual(yl, -math.acos(-0.5), 12)
        self.assertAlmostEqual(yu, -math.acos(0.5), 12)
        yl, yu = self.acos(-0.5, 0.5, -math.pi - 0.1, 0 + 0.1, 1e-8)
        self.assertAlmostEqual(yl, -math.acos(-0.5), 12)
        self.assertAlmostEqual(yu, -math.acos(0.5), 12)
        yl, yu = self.acos(-0.5, 0.5, -math.pi + 0.1, 0 - 0.1, 1e-8)
        self.assertAlmostEqual(yl, -math.acos(-0.5), 12)
        self.assertAlmostEqual(yu, -math.acos(0.5), 12)

    def test_atan(self):
        if not numpy_available:
            raise unittest.SkipTest('Numpy is not available.')
        yl, yu = self.atan(-0.5, 0.5, -interval.inf, interval.inf)
        self.assertEqual(yl, -interval.inf)
        self.assertEqual(yu, interval.inf)
        yl, yu = self.atan(-0.5, 0.5, -0.1, 0.1)
        self.assertAlmostEqual(yl, -0.1, 12)
        self.assertAlmostEqual(yu, 0.1, 12)
        yl, yu = self.atan(-0.5, 0.5, -0.5 * math.pi + 0.1, math.pi / 2 - 0.1)
        self.assertAlmostEqual(yl, math.atan(-0.5), 12)
        self.assertAlmostEqual(yu, math.atan(0.5), 12)
        yl, yu = self.atan(-0.5, 0.5, -1.5 * math.pi + 0.1, 1.5 * math.pi - 0.1)
        self.assertAlmostEqual(yl, math.atan(-0.5) - math.pi, 12)
        self.assertAlmostEqual(yu, math.atan(0.5) + math.pi, 12)

    def test_encountered_bugs(self):
        lb, ub = self._inverse_power1(
            88893.4225, 88893.4225, 2, 2, 298.15, 298.15, 1e-8
        )
        self.assertAlmostEqual(lb, 298.15)
        self.assertAlmostEqual(ub, 298.15)

        lb, ub = self._inverse_power1(2.56e-6, 2.56e-6, 2, 2, -0.0016, -0.0016, 1e-12)
        self.assertAlmostEqual(lb, -0.0016)
        self.assertAlmostEqual(ub, -0.0016)

        lb, ub = self._inverse_power1(
            -1, -1e-12, 2, 2, -interval.inf, interval.inf, 1e-8
        )
        self.assertAlmostEqual(lb, 0)
        self.assertAlmostEqual(ub, 0)

        lb, ub = self.mul(0, 0, -interval.inf, interval.inf)
        self.assertEqual(lb, -interval.inf)
        self.assertEqual(ub, interval.inf)


class TestInterval(IntervalTestBase, unittest.TestCase):
    def setUp(self):
        super(TestInterval, self).setUp()
        self.add = interval.add
        self.sub = interval.sub
        self.mul = interval.mul
        self.inv = interval.inv
        self.div = interval.div
        self.power = interval.power
        self.exp = interval.exp
        self.log = interval.log
        self.log10 = interval.log10
        self.sin = interval.sin
        self.cos = interval.cos
        self.tan = interval.tan
        self.asin = interval.asin
        self.acos = interval.acos
        self.atan = interval.atan
        self._inverse_power1 = interval._inverse_power1
        self._inverse_power2 = interval._inverse_power2



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Problem Writer for CPLEX LP Format Files
#

import logging

from pyomo.common.gc_manager import PauseGC
from pyomo.opt import ProblemFormat
from pyomo.opt.base import AbstractProblemWriter, WriterFactory
from pyomo.core.base import (
    SymbolMap,
    TextLabeler,
    NumericLabeler,
    Constraint,
    SortComponents,
    Var,
    value,
    SOSConstraint,
    Objective,
    ComponentMap,
    is_fixed,
)
from pyomo.repn import generate_standard_repn

logger = logging.getLogger('pyomo.core')


def _no_negative_zero(val):
    """Make sure -0 is never output. Makes diff tests easier."""
    if val == 0:
        return 0
    return val


def _get_bound(exp):
    if exp is None:
        return None
    if is_fixed(exp):
        return value(exp)
    raise ValueError("non-fixed bound or weight: " + str(exp))


@WriterFactory.register('cpxlp_v1', 'Generate the corresponding CPLEX LP file')
@WriterFactory.register('lp_v1', 'Generate the corresponding CPLEX LP file')
class ProblemWriter_cpxlp(AbstractProblemWriter):
    def __init__(self):
        AbstractProblemWriter.__init__(self, ProblemFormat.cpxlp)

        # The LP writer tracks which variables are
        # referenced in constraints, so that a user does not end up with a
        # zillion "unreferenced variables" warning messages.
        # This dictionary maps id(VarData) -> VarData.
        self._referenced_variable_ids = {}

        # Per ticket #4319, we are using %.17g, which mocks the
        # behavior of using %r (i.e., float('%r'%<number>) == <number>) with
        # the added benefit of outputting (+/-). The only case where this
        # fails to mock the behavior of %r is for large (long) integers (L),
        # which is a rare case to run into and is probably indicative of
        # other issues with the model.
        # *** NOTE ***: If you use 'r' or 's' here, it will break code that
        #               relies on using '%+' before the formatting character
        #               and you will need to go add extra logic to output
        #               the number's sign.
        self._precision_string = '.17g'
        self.linear_coef_string_template = '%+' + self._precision_string + ' %s\n'
        self.quad_coef_string_template = '%+' + self._precision_string + ' '
        self.obj_string_template = '%+' + self._precision_string + ' %s\n'
        self.sos_template_string = "%s:%" + self._precision_string + "\n"
        self.eq_string_template = "= %" + self._precision_string + '\n'
        self.geq_string_template = ">= %" + self._precision_string + '\n\n'
        self.leq_string_template = "<= %" + self._precision_string + '\n\n'
        self.lb_string_template = "%" + self._precision_string + " <= "
        self.ub_string_template = " <= %" + self._precision_string + "\n"

    def __call__(self, model, output_filename, solver_capability, io_options):
        # Make sure not to modify the user's dictionary,
        # they may be reusing it outside of this call
        io_options = dict(io_options)

        # Skip writing constraints whose body section is
        # fixed (i.e., no variables)
        skip_trivial_constraints = io_options.pop("skip_trivial_constraints", False)

        # Use full Pyomo component names in the LP file rather
        # than shortened symbols (slower, but useful for debugging).
        symbolic_solver_labels = io_options.pop("symbolic_solver_labels", False)

        output_fixed_variable_bounds = io_options.pop(
            "output_fixed_variable_bounds", False
        )

        # If False, unused variables will not be included in
        # the LP file. Otherwise, include all variables in
        # the bounds sections.
        include_all_variable_bounds = io_options.pop(
            "include_all_variable_bounds", False
        )

        labeler = io_options.pop("labeler", None)

        # Specify how much effort do we want to put into ensuring the
        # LP file is written deterministically for a Pyomo model:
        #    0 : None
        #    1 : sort keys of indexed components (default)
        #    2 : sort keys AND sort names (over declaration order)
        file_determinism = io_options.pop("file_determinism", 1)

        # Specify orderings for variable and constraint output
        row_order = io_options.pop("row_order", None)
        column_order = io_options.pop("column_order", None)

        # Make sure the ONE_VAR_CONSTANT variable appears in
        # the objective even if the constant part of the
        # objective is zero
        force_objective_constant = io_options.pop("force_objective_constant", False)

        if len(io_options):
            raise ValueError(
                "ProblemWriter_cpxlp passed unrecognized io_options:\n\t"
                + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
            )

        if symbolic_solver_labels and (labeler is not None):
            raise ValueError(
                "ProblemWriter_cpxlp: Using both the "
                "'symbolic_solver_labels' and 'labeler' "
                "I/O options is forbidden"
            )

        #
        # Create labeler
        #
        if symbolic_solver_labels:
            labeler = TextLabeler()
        elif labeler is None:
            labeler = NumericLabeler('x')

        # Clear the collection of referenced variables.
        self._referenced_variable_ids.clear()

        if output_filename is None:
            output_filename = model.name + ".lp"

        # when sorting, there are a non-trivial number of
        # temporary objects created. these all yield
        # non-circular references, so disable GC - the
        # overhead is non-trivial, and because references
        # are non-circular, everything will be collected
        # immediately anyway.
        with PauseGC() as pgc:
            with open(output_filename, "w") as output_file:
                symbol_map = self._print_model_LP(
                    model,
                    output_file,
                    solver_capability,
                    labeler,
                    output_fixed_variable_bounds=output_fixed_variable_bounds,
                    file_determinism=file_determinism,
                    row_order=row_order,
                    column_order=column_order,
                    skip_trivial_constraints=skip_trivial_constraints,
                    force_objective_constant=force_objective_constant,
                    include_all_variable_bounds=include_all_variable_bounds,
                )

        self._referenced_variable_ids.clear()

        return output_filename, symbol_map

    def _print_expr_canonical(
        self,
        x,
        output,
        object_symbol_dictionary,
        variable_symbol_dictionary,
        is_objective,
        column_order,
        file_determinism,
        force_objective_constant=False,
    ):
        """
        Return a expression as a string in LP format.

        Note that this function does not handle any differences in LP format
        interpretation by the solvers (e.g. CPlex vs GLPK).  That decision is
        left up to the caller.

        required arguments:
          x: A Pyomo canonical expression to write in LP format
        """
        assert (not force_objective_constant) or (is_objective)
        linear_coef_string_template = self.linear_coef_string_template
        quad_coef_string_template = self.quad_coef_string_template

        constant = True
        #
        # Linear
        #
        if len(x.linear_vars) > 0:
            constant = False
            for vardata in x.linear_vars:
                self._referenced_variable_ids[id(vardata)] = vardata

            if column_order is None:
                #
                # Order columns by dictionary names
                #
                names = [variable_symbol_dictionary[id(var)] for var in x.linear_vars]

                term_iterator = zip(x.linear_coefs, names)
                if file_determinism > 0:
                    term_iterator = sorted(term_iterator, key=lambda x: x[1])

                for coef, name in term_iterator:
                    output.append(linear_coef_string_template % (coef, name))
            else:
                #
                # Order columns by the value of column_order[]
                #
                for i, var in sorted(
                    enumerate(x.linear_vars), key=lambda x: column_order[x[1]]
                ):
                    name = variable_symbol_dictionary[id(var)]
                    output.append(
                        linear_coef_string_template % (x.linear_coefs[i], name)
                    )

        #
        # Quadratic
        #
        if len(x.quadratic_vars) > 0:
            constant = False
            for var1, var2 in x.quadratic_vars:
                self._referenced_variable_ids[id(var1)] = var1
                self._referenced_variable_ids[id(var2)] = var2

            output.append("+ [\n")

            if column_order is None:
                #
                # Order columns by dictionary names
                #
                quad = set()
                names = []
                i = 0
                for var1, var2 in x.quadratic_vars:
                    name1 = variable_symbol_dictionary[id(var1)]
                    name2 = variable_symbol_dictionary[id(var2)]
                    if name1 < name2:
                        names.append((name1, name2))
                    elif name1 > name2:
                        names.append((name2, name1))
                    else:
                        quad.add(i)
                        names.append((name1, name1))
                    i += 1

                term_iterator = enumerate(names)
                if file_determinism > 0:
                    term_iterator = sorted(term_iterator, key=lambda x: x[1])
                for i, names_ in term_iterator:
                    #
                    # Times 2 because LP format requires /2 for all the quadratic
                    # terms /of the objective only/.  Discovered the last bit thru
                    # trial and error.  Obnoxious.
                    # Ref: ILog CPlex 8.0 User's Manual, p197.
                    #
                    if is_objective:
                        tmp = 2 * x.quadratic_coefs[i]
                        output.append(quad_coef_string_template % tmp)
                    else:
                        output.append(quad_coef_string_template % x.quadratic_coefs[i])
                    if i in quad:
                        output.append("%s ^ 2\n" % (names_[0]))
                    else:
                        output.append("%s * %s\n" % (names_[0], names_[1]))
            else:
                #
                # Order columns by the value of column_order[]
                #
                quad = set()
                cols = []
                i = 0
                for var1, var2 in x.quadratic_vars:
                    col1 = column_order[var1]
                    col2 = column_order[var2]
                    if col1 < col2:
                        cols.append(
                            (
                                (
                                    (col1, col2),
                                    variable_symbol_dictionary[id(var1)],
                                    variable_symbol_dictionary[id(var2)],
                                )
                            )
                        )
                    elif col1 > col2:
                        cols.append(
                            (
                                (
                                    (col2, col1),
                                    variable_symbol_dictionary[id(var2)],
                                    variable_symbol_dictionary[id(var1)],
                                )
                            )
                        )
                    else:
                        quad.add(i)
                        cols.append(
                            ((col1, col1), variable_symbol_dictionary[id(var1)])
                        )
                    i += 1
                for i, cols_ in sorted(enumerate(cols), key=lambda x: x[1][0]):
                    #
                    # Times 2 because LP format requires /2 for all the quadratic
                    # terms /of the objective only/.  Discovered the last bit thru
                    # trial and error.  Obnoxious.
                    # Ref: ILog CPlex 8.0 User's Manual, p197.
                    #
                    if is_objective:
                        output.append(
                            quad_coef_string_template % 2 * x.quadratic_coefs[i]
                        )
                    else:
                        output.append(quad_coef_string_template % x.quadratic_coefs[i])
                    if i in quad:
                        output.append("%s ^ 2\n" % cols_[1])
                    else:
                        output.append("%s * %s\n" % (cols_[1], cols_[2]))

            output.append("]")

            if is_objective:
                output.append(' / 2\n')
                # divide by 2 because LP format requires /2 for all the quadratic
                # terms.  Weird.  Ref: ILog CPlex 8.0 User's Manual, p197
            else:
                output.append("\n")

        if constant and not is_objective:
            # If we made it to here we are outputting
            # trivial constraints place 0 *
            # ONE_VAR_CONSTANT on this side of the
            # constraint for the benefit of solvers like
            # Glpk that cannot parse an LP file without
            # a variable on the left hand side.
            output.append(linear_coef_string_template % (0, 'ONE_VAR_CONSTANT'))

        #
        # Constant offset
        #
        # Currently, it appears that we only need to print
        # the constant offset term for objectives.
        #
        if is_objective and (force_objective_constant or (x.constant != 0.0)):
            output.append(self.obj_string_template % (x.constant, 'ONE_VAR_CONSTANT'))

        #
        # Return constant offset
        #
        return x.constant

    def printSOS(self, symbol_map, labeler, variable_symbol_map, soscondata, output):
        """
        Prints the SOS constraint associated with the SOSConstraintData object
        """
        sos_template_string = self.sos_template_string

        if hasattr(soscondata, 'get_items'):
            sos_items = list(soscondata.get_items())
        else:
            sos_items = list(soscondata.items())

        if len(sos_items) == 0:
            return

        level = soscondata.level

        output.append(
            '%s: S%s::\n' % (symbol_map.getSymbol(soscondata, labeler), level)
        )

        for vardata, weight in sos_items:
            weight = _get_bound(weight)
            if weight < 0:
                raise ValueError(
                    "Cannot use negative weight %f "
                    "for variable %s is special ordered "
                    "set %s " % (weight, vardata.name, soscondata.name)
                )
            if vardata.fixed:
                raise RuntimeError(
                    "SOSConstraint '%s' includes a fixed variable '%s'. This is "
                    "currently not supported. Deactivate this constraint in order to "
                    "proceed." % (soscondata.name, vardata.name)
                )
            self._referenced_variable_ids[id(vardata)] = vardata
            output.append(
                sos_template_string % (variable_symbol_map.getSymbol(vardata), weight)
            )

    def _print_model_LP(
        self,
        model,
        output_file,
        solver_capability,
        labeler,
        output_fixed_variable_bounds=False,
        file_determinism=1,
        row_order=None,
        column_order=None,
        skip_trivial_constraints=False,
        force_objective_constant=False,
        include_all_variable_bounds=False,
    ):
        eq_string_template = self.eq_string_template
        leq_string_template = self.leq_string_template
        geq_string_template = self.geq_string_template
        ub_string_template = self.ub_string_template
        lb_string_template = self.lb_string_template

        symbol_map = SymbolMap()
        variable_symbol_map = SymbolMap()
        # NOTE: we use createSymbol instead of getSymbol because we
        #       know whether or not the symbol exists, and don't want
        #       to the overhead of error/duplicate checking.
        # cache frequently called functions
        create_symbol_func = SymbolMap.createSymbol
        create_symbols_func = SymbolMap.createSymbols
        alias_symbol_func = SymbolMap.alias
        variable_label_pairs = []

        # populate the symbol map in a single pass.
        # objective_list, constraint_list, sosconstraint_list, variable_list \
        #    = self._populate_symbol_map(model,
        #                                symbol_map,
        #                                labeler,
        #                                variable_symbol_map,
        #                                file_determinism=file_determinism)
        sortOrder = SortComponents.unsorted
        if file_determinism >= 1:
            sortOrder = sortOrder | SortComponents.indices
            if file_determinism >= 2:
                sortOrder = sortOrder | SortComponents.alphabetical

        #
        # Create variable symbols (and cache the block list)
        #
        all_blocks = []
        variable_list = []
        #
        # WEH - TODO:  See if this is faster
        # NOTE: This loop doesn't find all of the variables.  :(
        #
        # for block in model.block_data_objects(active=True,
        #                                      sort=sortOrder):
        #
        #    all_blocks.append(block)
        #
        #    for vardata in block.component_data_objects(
        #            Var,
        #            active=True,
        #            sort=sortOrder,
        #            descend_into=False):
        #
        #        variable_list.append(vardata)
        #        variable_label_pairs.append(
        #            (vardata,create_symbol_func(symbol_map,
        #                                        vardata,
        #                                        labeler)))
        all_blocks = list(model.block_data_objects(active=True, sort=sortOrder))
        variable_list = list(model.component_data_objects(Var, sort=sortOrder))
        variable_label_pairs = list(
            (vardata, create_symbol_func(symbol_map, vardata, labeler))
            for vardata in variable_list
        )
        variable_symbol_map.addSymbols(variable_label_pairs)

        # and extract the information we'll need for rapid labeling.
        object_symbol_dictionary = symbol_map.byObject
        variable_symbol_dictionary = variable_symbol_map.byObject

        # TODO: this is a quick wrapper to map a very unintuitive
        # exception to a more meaningful error.  Future versions of the
        # LP writer may be more accepting in what expressions/variables
        # they accept.
        def print_expr_canonical(
            obj,
            x,
            output,
            object_symbol_dictionary,
            variable_symbol_dictionary,
            is_objective,
            column_order,
            file_determinism,
            force_objective_constant=False,
        ):
            try:
                return self._print_expr_canonical(
                    x=x,
                    output=output,
                    object_symbol_dictionary=object_symbol_dictionary,
                    variable_symbol_dictionary=variable_symbol_dictionary,
                    is_objective=is_objective,
                    column_order=column_order,
                    file_determinism=file_determinism,
                    force_objective_constant=force_objective_constant,
                )
            except KeyError as e:
                _id = e.args[0]
                _var = None
                if x.linear_vars:
                    for v in x.linear_vars:
                        if id(v) == _id:
                            _var = v
                            break
                if _var is None and x.quadratic_vars:
                    for v in x.quadratic_vars:
                        v = [_v for _v in v if id(_v) == _id]
                        if v:
                            _var = v[0]
                            break
                if _var is not None:
                    logger.error(
                        "Model contains an expression (%s) that contains "
                        "a variable (%s) that is not attached to an active "
                        "block on the submodel being written" % (obj.name, _var.name)
                    )
                raise

        # print the model name and the source, so we know roughly where
        # it came from.
        #
        # NOTE: this *must* use the "\* ... *\" comment format: the GLPK
        # LP parser does not correctly handle other formats (notably, "%").
        output = []
        output.append("\\* Source Pyomo model name=%s *\\\n\n" % (model.name,))

        #
        # Objective
        #

        supports_quadratic_objective = solver_capability('quadratic_objective')

        numObj = 0
        onames = []
        for block in all_blocks:
            gen_obj_repn = getattr(block, "_gen_obj_repn", None)
            if gen_obj_repn is not None:
                gen_obj_repn = bool(gen_obj_repn)
                # Get/Create the ComponentMap for the repn
                if not hasattr(block, '_repn'):
                    block._repn = ComponentMap()
                block_repn = block._repn

            for objective_data in block.component_data_objects(
                Objective, active=True, sort=sortOrder, descend_into=False
            ):
                numObj += 1
                onames.append(objective_data.name)
                if numObj > 1:
                    raise ValueError(
                        "More than one active objective defined for input "
                        "model '%s'; Cannot write legal LP file\n"
                        "Objectives: %s" % (model.name, ' '.join(onames))
                    )

                create_symbol_func(symbol_map, objective_data, labeler)

                symbol_map.alias(objective_data, '__default_objective__')
                if objective_data.is_minimizing():
                    output.append("min \n")
                else:
                    output.append("max \n")

                if gen_obj_repn == False:
                    repn = block_repn[objective_data]
                else:
                    repn = generate_standard_repn(
                        objective_data.expr, quadratic=supports_quadratic_objective
                    )
                    if gen_obj_repn:
                        block_repn[objective_data] = repn

                degree = repn.polynomial_degree()

                if degree == 0:
                    logger.warning(
                        "Constant objective detected, replacing "
                        "with a placeholder to prevent solver failure."
                    )
                    force_objective_constant = True
                elif degree == 2:
                    if not supports_quadratic_objective:
                        raise RuntimeError(
                            "Selected solver is unable to handle "
                            "objective functions with quadratic terms. "
                            "Objective at issue: %s." % objective_data.name
                        )
                elif degree is None:
                    raise RuntimeError(
                        "Cannot write legal LP file.  Objective '%s' "
                        "has nonlinear terms that are not quadratic."
                        % objective_data.name
                    )

                output.append(object_symbol_dictionary[id(objective_data)] + ':\n')

                offset = print_expr_canonical(
                    objective_data,
                    repn,
                    output,
                    object_symbol_dictionary,
                    variable_symbol_dictionary,
                    True,
                    column_order,
                    file_determinism,
                    force_objective_constant=force_objective_constant,
                )

        if numObj == 0:
            raise ValueError(
                "ERROR: No objectives defined for input model. "
                "Cannot write legal LP file."
            )

        # Constraints
        #
        # If there are no non-trivial constraints, you'll end up with an empty
        # constraint block. CPLEX is OK with this, but GLPK isn't. And
        # eliminating the constraint block (i.e., the "s.t." line) causes GLPK
        # to whine elsewhere. Output a warning if the constraint block is empty,
        # so users can quickly determine the cause of the solve failure.

        output.append("\n")
        output.append("s.t.\n")
        output.append("\n")

        have_nontrivial = False

        supports_quadratic_constraint = solver_capability('quadratic_constraint')

        def constraint_generator():
            for block in all_blocks:
                gen_con_repn = getattr(block, "_gen_con_repn", None)
                if gen_con_repn is not None:
                    gen_con_repn = bool(gen_con_repn)
                    # Get/Create the ComponentMap for the repn
                    if not hasattr(block, '_repn'):
                        block._repn = ComponentMap()
                    block_repn = block._repn

                for constraint_data in block.component_data_objects(
                    Constraint, active=True, sort=sortOrder, descend_into=False
                ):
                    if (not constraint_data.has_lb()) and (
                        not constraint_data.has_ub()
                    ):
                        assert not constraint_data.equality
                        continue  # non-binding, so skip

                    if gen_con_repn == False:
                        repn = block_repn[constraint_data]
                    else:
                        if constraint_data._linear_canonical_form:
                            repn = constraint_data.canonical_form()
                        else:
                            repn = generate_standard_repn(
                                constraint_data.body,
                                quadratic=supports_quadratic_constraint,
                            )
                        if gen_con_repn:
                            block_repn[constraint_data] = repn

                    yield constraint_data, repn

        if row_order is not None:
            sorted_constraint_list = list(constraint_generator())
            sorted_constraint_list.sort(key=lambda x: row_order[x[0]])

            def yield_all_constraints():
                yield from sorted_constraint_list

        else:
            yield_all_constraints = constraint_generator

        # FIXME: This is a hack to get nested blocks working...
        for constraint_data, repn in yield_all_constraints():
            have_nontrivial = True

            degree = repn.polynomial_degree()

            #
            # Write constraint
            #

            # There are conditions, e.g., when fixing variables, under which
            # a constraint block might be empty.  Ignore these, for both
            # practical reasons and the fact that the CPLEX LP format
            # requires a variable in the constraint body.  It is also
            # possible that the body of the constraint consists of only a
            # constant, in which case the "variable" of
            if degree == 0:
                if skip_trivial_constraints:
                    continue
            elif degree == 2:
                if not supports_quadratic_constraint:
                    raise ValueError(
                        "Solver unable to handle quadratic expressions. Constraint"
                        " at issue: '%s'" % (constraint_data.name)
                    )
            elif degree is None:
                raise ValueError(
                    "Cannot write legal LP file.  Constraint '%s' has a body "
                    "with nonlinear terms." % (constraint_data.name)
                )

            # Create symbol
            con_symbol = create_symbol_func(symbol_map, constraint_data, labeler)

            if constraint_data.equality:
                assert value(constraint_data.lower) == value(constraint_data.upper)
                label = 'c_e_%s_' % con_symbol
                alias_symbol_func(symbol_map, constraint_data, label)
                output.append(label)
                output.append(':\n')
                offset = print_expr_canonical(
                    constraint_data,
                    repn,
                    output,
                    object_symbol_dictionary,
                    variable_symbol_dictionary,
                    False,
                    column_order,
                    file_determinism,
                )
                bound = constraint_data.lower
                bound = _get_bound(bound) - offset
                output.append(eq_string_template % (_no_negative_zero(bound)))
                output.append("\n")
            else:
                if constraint_data.has_lb():
                    if constraint_data.has_ub():
                        label = 'r_l_%s_' % con_symbol
                    else:
                        label = 'c_l_%s_' % con_symbol
                    alias_symbol_func(symbol_map, constraint_data, label)
                    output.append(label)
                    output.append(':\n')
                    offset = print_expr_canonical(
                        constraint_data,
                        repn,
                        output,
                        object_symbol_dictionary,
                        variable_symbol_dictionary,
                        False,
                        column_order,
                        file_determinism,
                    )
                    bound = constraint_data.lower
                    bound = _get_bound(bound) - offset
                    output.append(geq_string_template % (_no_negative_zero(bound)))
                else:
                    assert constraint_data.has_ub()

                if constraint_data.has_ub():
                    if constraint_data.has_lb():
                        label = 'r_u_%s_' % con_symbol
                    else:
                        label = 'c_u_%s_' % con_symbol
                    alias_symbol_func(symbol_map, constraint_data, label)
                    output.append(label)
                    output.append(':\n')
                    offset = print_expr_canonical(
                        constraint_data,
                        repn,
                        output,
                        object_symbol_dictionary,
                        variable_symbol_dictionary,
                        False,
                        column_order,
                        file_determinism,
                    )
                    bound = constraint_data.upper
                    bound = _get_bound(bound) - offset
                    output.append(leq_string_template % (_no_negative_zero(bound)))
                else:
                    assert constraint_data.has_lb()

            # A simple hack to avoid caching super large files
            if len(output) > 1024:
                output_file.write("".join(output))
                output = []

        if not have_nontrivial:
            logger.warning(
                'Empty constraint block written in LP format - solver may error'
            )

        # the CPLEX LP format doesn't allow constants in the objective (or
        # constraint body), which is a bit silly.  To avoid painful
        # book-keeping, we introduce the following "variable", constrained
        # to the value 1.  This is used when quadratic terms are present.
        # worst-case, if not used, is that CPLEX easily pre-processes it out.
        prefix = ""
        output.append('%sc_e_ONE_VAR_CONSTANT: \n' % prefix)
        output.append('%sONE_VAR_CONSTANT = 1.0\n' % prefix)
        output.append("\n")

        # SOS constraints
        #
        # For now, we write out SOS1 and SOS2 constraints in the cplex format
        #
        # All Component objects are stored in model._component, which is a
        # dictionary of {class: {objName: object}}.
        #
        # Consider the variable X,
        #
        #   model.X = Var(...)
        #
        # We print X to CPLEX format as X(i,j,k,...) where i, j, k, ... are the
        # indices of X.
        #
        SOSlines = []
        sos1 = solver_capability("sos1")
        sos2 = solver_capability("sos2")
        writtenSOS = False
        for block in all_blocks:
            for soscondata in block.component_data_objects(
                SOSConstraint, active=True, sort=sortOrder, descend_into=False
            ):
                create_symbol_func(symbol_map, soscondata, labeler)

                level = soscondata.level
                if (
                    (level == 1 and not sos1)
                    or (level == 2 and not sos2)
                    or (level > 2)
                ):
                    raise ValueError(
                        "Solver does not support SOS level %s constraints" % (level)
                    )
                if writtenSOS == False:
                    SOSlines.append("SOS\n")
                    writtenSOS = True
                # This updates the referenced_variable_ids, just in case
                # there is a variable that only appears in an
                # SOSConstraint, in which case this needs to be known
                # before we write the "bounds" section (Cplex does not
                # handle this correctly, Gurobi does)
                self.printSOS(
                    symbol_map, labeler, variable_symbol_map, soscondata, SOSlines
                )

        #
        # Bounds
        #

        output.append("bounds\n")

        # Scan all variables even if we're only writing a subset of them.
        # required because we don't store maps by variable type currently.

        # Track the number of integer and binary variables, so you can
        # output their status later.
        integer_vars = []
        binary_vars = []
        for vardata in variable_list:
            # TODO: We could just loop over the set of items in
            #       self._referenced_variable_ids, except this is
            #       a dictionary that is hashed by id(vardata)
            #       which would make the bounds section
            #       nondeterministic (bad for unit testing)
            if (not include_all_variable_bounds) and (
                id(vardata) not in self._referenced_variable_ids
            ):
                continue

            name_to_output = variable_symbol_dictionary[id(vardata)]
            if name_to_output == "e":
                raise ValueError(
                    "Attempting to write variable with name 'e' in a CPLEX LP "
                    "formatted file will cause a parse failure due to confusion with "
                    "numeric values expressed in scientific notation"
                )

            # track the number of integer and binary variables, so we know whether
            # to output the general / binary sections below.
            if vardata.is_binary():
                binary_vars.append(name_to_output)
            elif vardata.is_integer():
                integer_vars.append(name_to_output)
            elif not vardata.is_continuous():
                raise TypeError(
                    "Invalid domain type for variable with name '%s'. "
                    "Variable is not continuous, integer, or binary." % (vardata.name)
                )

            if vardata.fixed:
                if not output_fixed_variable_bounds:
                    raise ValueError(
                        "Encountered a fixed variable (%s) inside an active "
                        "objective or constraint expression on model %s, which is "
                        "usually indicative of a preprocessing error. Use the "
                        "IO-option 'output_fixed_variable_bounds=True' to suppress "
                        "this error and fix the variable by overwriting its bounds "
                        "in the LP file." % (vardata.name, model.name)
                    )
                if vardata.value is None:
                    raise ValueError("Variable cannot be fixed to a value of None.")
                vardata_lb = value(vardata.value)
                vardata_ub = value(vardata.value)

                output.append("   ")
                output.append(lb_string_template % (_no_negative_zero(vardata_lb)))
                output.append(name_to_output)
                output.append(ub_string_template % (_no_negative_zero(vardata_ub)))
            else:
                vardata_lb = _get_bound(vardata.lb)
                vardata_ub = _get_bound(vardata.ub)

                # Pyomo assumes that the default variable bounds are -inf and +inf
                output.append("   ")
                if vardata.has_lb():
                    output.append(lb_string_template % (_no_negative_zero(vardata_lb)))
                else:
                    output.append(" -inf <= ")

                output.append(name_to_output)
                if vardata.has_ub():
                    output.append(ub_string_template % (_no_negative_zero(vardata_ub)))
                else:
                    output.append(" <= +inf\n")

        if len(integer_vars) > 0:
            output.append("general\n")
            for var_name in integer_vars:
                output.append('  %s\n' % var_name)

        if len(binary_vars) > 0:
            output.append("binary\n")
            for var_name in binary_vars:
                output.append('  %s\n' % var_name)

        # Write the SOS section
        output.append("".join(SOSlines))

        #
        # wrap-up
        #
        output.append("end\n")
        output_file.write("".join(output))

        # Clean up the symbol map to only contain variables referenced
        # in the active constraints **Note**: warm start method may
        # rely on this for choosing the set of potential warm start
        # variables
        vars_to_delete = set(variable_symbol_map.byObject.keys()) - set(
            self._referenced_variable_ids.keys()
        )
        sm_byObject = symbol_map.byObject
        sm_bySymbol = symbol_map.bySymbol
        var_sm_byObject = variable_symbol_map.byObject
        for varid in vars_to_delete:
            symbol = var_sm_byObject[varid]
            del sm_byObject[varid]
            del sm_bySymbol[symbol]
        del variable_symbol_map

        return symbol_map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import os
import stat
import sys
from pyomo.common.download import FileDownloader

logger = logging.getLogger('pyomo.common')

# These URLs were retrieved from
#     https://dev.ampl.com/more/reports.html#hooking-solvers-to-ampl
# All 32-bit downloads are used - 64-bit is available only for Linux
urlmap = {
    'linux': 'https://netlib.org/ampl/student/linux/gjh.gz',
    'windows': 'https://netlib.org/ampl/student/mswin/gjh.exe.gz',
    'cygwin': 'https://netlib.org/ampl/student/mswin/gjh.exe.gz',
    'darwin': 'https://netlib.org/ampl/student/macosx/x86_32/gjh.gz',
}
exemap = {'linux': '', 'windows': '.exe', 'cygwin': '.exe', 'darwin': ''}


def get_gjh(downloader):
    system, bits = downloader.get_sysinfo()
    url = downloader.get_platform_url(urlmap)

    downloader.set_destination_filename(os.path.join('bin', 'gjh' + exemap[system]))

    logger.info(
        "Fetching GJH from %s and installing it to %s" % (url, downloader.destination())
    )

    downloader.get_gzipped_binary_file(url)

    mode = os.stat(downloader.destination()).st_mode
    os.chmod(
        downloader.destination(), mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )


def main(argv):
    downloader = FileDownloader()
    downloader.parse_args(argv)
    get_gjh(downloader)


if __name__ == '__main__':
    logger.setLevel(logging.INFO)
    try:
        main(sys.argv[1:])
    except Exception as e:
        print(e.message)
        print("Usage: %s [--insecure] [target]" % os.path.basename(sys.argv[0]))
        sys.exit(1)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from contextlib import nullcontext

from pyomo.common import enums
from pyomo.common.deprecation import deprecated
from pyomo.common.modeling import NOTSET

TO_STRING_VERBOSE = False

# logical propositions
_and = 0
_or = 1
_inv = 2
_equiv = 3
_xor = 4
_impl = 5


#
# Provide a global value that indicates which expression system is being used
#
class Mode(enums.IntEnum):
    # coopr: Original Coopr/Pyomo expression system
    coopr_trees = 1
    # coopr3: leverage reference counts to reduce the amount of required
    # expression cloning to ensure independent expression trees.
    coopr3_trees = 3
    # pyomo4: rework the expression system to remove reliance on
    # reference counting.  This enables pypy support (which doesn't have
    # reference counting).  This version never became the default.
    pyomo4_trees = 4
    # pyomo5: refinement of pyomo4.  Expressions are now immutable by
    # contract, which tolerates "entangled" expression trees.  Added
    # specialized classes for NPV expressions and LinearExpressions.
    pyomo5_trees = 5
    # pyomo6: refinement of pyomo5 expression generation to leverage
    # multiple dispatch.  Standardized expression storage and argument
    # handling (significant rework of the LinearExpression structure).
    pyomo6_trees = 6
    #
    CURRENT = pyomo6_trees


_mode = Mode.CURRENT
# We no longer support concurrent expression systems.  _mode is left
# primarily so we can support expression system-specific baselines
assert _mode == Mode.pyomo6_trees


class OperatorAssociativity(enums.IntEnum):
    """Enum for indicating the associativity of an operator.

    LEFT_TO_RIGHT(1) if this operator is left-to-right associative or
    RIGHT_TO_LEFT(-1) if it is right-to-left associative.  Any other
    values will be interpreted as "not associative" (implying any
    arguments that are at this operator's PRECEDENCE will be enclosed
    in parens).

    """

    RIGHT_TO_LEFT = -1
    NON_ASSOCIATIVE = 0
    LEFT_TO_RIGHT = 1


class ExpressionType(enums.Enum):
    NUMERIC = 0
    RELATIONAL = 1
    LOGICAL = 2


class NUMERIC_ARG_TYPE(enums.IntEnum):
    MUTABLE = -2
    ASNUMERIC = -1
    INVALID = 0
    NATIVE = 1
    NPV = 2
    PARAM = 3
    VAR = 4
    MONOMIAL = 5
    LINEAR = 6
    SUM = 7
    OTHER = 8


class RELATIONAL_ARG_TYPE(enums.IntEnum, metaclass=enums.ExtendedEnumType):
    __base_enum__ = NUMERIC_ARG_TYPE

    INEQUALITY = 100
    INVALID_RELATIONAL = 101


def _invalid(*args):
    return NotImplemented


def _recast_mutable(expr):
    expr.make_immutable()
    if expr._nargs > 1:
        return expr
    elif not expr._nargs:
        return 0
    else:
        return expr._args_[0]


def _unary_op_dispatcher_type_mapping(dispatcher, updates, TYPES=NUMERIC_ARG_TYPE):
    #
    # Special case (wrapping) operators
    #
    def _asnumeric(a):
        a = a.as_numeric()
        return dispatcher[a.__class__](a)

    def _mutable(a):
        a = _recast_mutable(a)
        return dispatcher[a.__class__](a)

    mapping = {
        TYPES.ASNUMERIC: _asnumeric,
        TYPES.MUTABLE: _mutable,
        TYPES.INVALID: _invalid,
    }

    mapping.update(updates)
    return mapping


def _binary_op_dispatcher_type_mapping(dispatcher, updates, TYPES=NUMERIC_ARG_TYPE):
    #
    # Special case (wrapping) operators
    #
    def _any_asnumeric(a, b):
        b = b.as_numeric()
        return dispatcher[a.__class__, b.__class__](a, b)

    def _asnumeric_any(a, b):
        a = a.as_numeric()
        return dispatcher[a.__class__, b.__class__](a, b)

    def _asnumeric_asnumeric(a, b):
        a = a.as_numeric()
        b = b.as_numeric()
        return dispatcher[a.__class__, b.__class__](a, b)

    def _any_mutable(a, b):
        b = _recast_mutable(b)
        return dispatcher[a.__class__, b.__class__](a, b)

    def _mutable_any(a, b):
        a = _recast_mutable(a)
        return dispatcher[a.__class__, b.__class__](a, b)

    def _mutable_mutable(a, b):
        if a is b:
            # Note: _recast_mutable is an in-place operation: make sure
            # that we don't call it twice on the same object.
            a = b = _recast_mutable(a)
        else:
            a = _recast_mutable(a)
            b = _recast_mutable(b)
        return dispatcher[a.__class__, b.__class__](a, b)

    mapping = {}

    # Because ASNUMERIC and MUTABLE re-call the dispatcher, we want to
    # resolve ASNUMERIC first, MUTABLE second, and INVALID last.  That
    # means we will add them to the dispatcher dict in opposite order so
    # "higher priority" callbacks override lower priority ones.

    mapping.update({(i, TYPES.INVALID): _invalid for i in TYPES})
    mapping.update({(TYPES.INVALID, i): _invalid for i in TYPES})

    mapping.update({(i, TYPES.MUTABLE): _any_mutable for i in TYPES})
    mapping.update({(TYPES.MUTABLE, i): _mutable_any for i in TYPES})
    mapping[TYPES.MUTABLE, TYPES.MUTABLE] = _mutable_mutable

    mapping.update({(i, TYPES.ASNUMERIC): _any_asnumeric for i in TYPES})
    mapping.update({(TYPES.ASNUMERIC, i): _asnumeric_any for i in TYPES})
    mapping[TYPES.ASNUMERIC, TYPES.ASNUMERIC] = _asnumeric_asnumeric

    mapping.update(updates)
    return mapping


@deprecated(
    """The clone counter has been removed and will always return 0.

Beginning with Pyomo5 expressions, expression cloning (detangling) no
longer occurs automatically within expression generation.  As a result,
the 'clone counter' has lost its utility and is no longer supported.
This context manager will always report 0.""",
    version='6.4.3',
)
class clone_counter(nullcontext):
    """Context manager for counting cloning events.

    This context manager counts the number of times that the
    :func:`clone_expression <pyomo.core.expr.current.clone_expression>`
    function is executed.
    """

    _count = 0

    def __init__(self):
        super().__init__(enter_result=self)

    @property
    def count(self):
        """A property that returns the clone count value."""
        return clone_counter._count


def _type_check_exception_arg(cls, exception):
    if exception is NOTSET:
        return True
    elif type(exception) is not bool:
        raise ValueError(
            f"{cls.ctype.__name__} '{cls.name}' was called with a non-bool "
            f"argument for 'exception': {exception}"
        )
    else:
        return exception
